import autoarray as aa
from autoarray.util import plotter_util
from autoastro.plotters import lens_plotter_util


def image(
    light_profile,
    grid,
    mask=None,
    positions=None,
    as_subplot=False,
    unit_label="arcsec",
    unit_conversion_factor=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Image",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    position_pointsize=10,
    grid_pointsize=1,
    output_path=None,
    output_format="show",
    output_filename="image",
):
    """Plot the image of a light profile, on a grid of (y,x) coordinates.

    Set *autoastro.hyper_galaxies.arrays.plotters.array_plotters* for a description of all innput parameters not described below.

    Parameters
    -----------
    light_profile : model.profiles.light_profiles.LightProfile
        The light profile whose image are plotted.
    grid : ndarray or hyper_galaxies.arrays.grid_stacks.Grid
        The (y,x) coordinates of the grid, in an arrays of shape (total_coordinates, 2)
    """
    profile_image = light_profile.profile_image_from_grid(grid=grid)

    aa.plot.array(
        array=profile_image,
        mask=mask,
        points=positions,
        as_subplot=as_subplot,
        unit_label=unit_label,
        unit_conversion_factor=unit_conversion_factor,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        point_pointsize=position_pointsize,
        grid_pointsize=grid_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def luminosity_within_circle_in_electrons_per_second_as_function_of_radius(
    light_profile,
    minimum_radius=1.0e-4,
    maximum_radius=10.0,
    radii_bins=10,
    as_subplot=False,
    label="Light Profile",
    plot_axis_type="semilogy",
    effective_radius_line=None,
    einstein_radius_line=None,
    unit_label="arcsec",
    unit_conversion_factor=None,
    figsize=(7, 7),
    plot_legend=True,
    title="Luminosity (Electrons Per Second) vs Radius",
    ylabel="Luminosity (Electrons Per Second)",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    legend_fontsize=12,
    output_path=None,
    output_format="show",
    output_filename="luminosity_vs_radius",
):

    radii = plotter_util.quantity_radii_from_minimum_and_maximum_radii_and_radii_points(
        minimum_radius=minimum_radius,
        maximum_radius=maximum_radius,
        radii_points=radii_bins,
    )

    luminosities = list(
        map(
            lambda radius: light_profile.luminosity_within_circle_in_units(
                radius=radius
            ),
            radii,
        )
    )

    aa.plot.quantity_as_function_of_radius(
        quantity=luminosities,
        radii=radii,
        as_subplot=as_subplot,
        label=label,
        plot_axis_type=plot_axis_type,
        effective_radius_line=effective_radius_line,
        einstein_radius_line=einstein_radius_line,
        unit_label=unit_label,
        unit_conversion_factor=unit_conversion_factor,
        figsize=figsize,
        plot_legend=plot_legend,
        title=title,
        ylabel=ylabel,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        legend_fontsize=legend_fontsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def convergence(
    mass_profile,
    grid,
    mask=None,
    positions=None,
    include_critical_curves=False,
    include_caustics=False,
    as_subplot=False,
    unit_label="arcsec",
    unit_conversion_factor=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Convergence",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    position_pointsize=10,
    grid_pointsize=1,
    output_path=None,
    output_format="show",
    output_filename="convergence",
):
    """Plot the convergence of a mass profile, on a grid of (y,x) coordinates.

    Set *autoastro.hyper_galaxies.arrays.plotters.array_plotters* for a description of all innput parameters not described below.

    Parameters
    -----------
    mass_profile : model.profiles.mass_profiles.MassProfile
        The mass profile whose convergence is plotted.
    grid : ndarray or hyper_galaxies.arrays.grid_stacks.Grid
        The (y,x) coordinates of the grid, in an arrays of shape (total_coordinates, 2)
    """

    convergence = mass_profile.convergence_from_grid(grid=grid)

    lines = lens_plotter_util.get_critical_curves_and_caustics_from_lensing_object(
        obj=mass_profile,
        include_critical_curves=include_critical_curves,
        include_caustics=include_caustics,
    )

    aa.plot.array(
        array=convergence,
        mask=mask,
        points=positions,
        lines=lines,
        as_subplot=as_subplot,
        unit_label=unit_label,
        unit_conversion_factor=unit_conversion_factor,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        point_pointsize=position_pointsize,
        grid_pointsize=grid_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def potential(
    mass_profile,
    grid,
    mask=None,
    positions=None,
    as_subplot=False,
    include_critical_curves=False,
    include_caustics=False,
    unit_label="arcsec",
    unit_conversion_factor=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Potential",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    position_pointsize=10,
    grid_pointsize=1,
    output_path=None,
    output_format="show",
    output_filename="potential",
):
    """Plot the potential of a mass profile, on a grid of (y,x) coordinates.

    Set *autoastro.hyper_galaxies.arrays.plotters.array_plotters* for a description of all innput parameters not described below.

    Parameters
    -----------
    mass_profile : model.profiles.mass_profiles.MassProfile
        The mass profile whose potential is plotted.
    grid : ndarray or hyper_galaxies.arrays.grid_stacks.Grid
        The (y,x) coordinates of the grid, in an arrays of shape (total_coordinates, 2)
    """
    potential = mass_profile.potential_from_grid(grid=grid)

    lines = lens_plotter_util.get_critical_curves_and_caustics_from_lensing_object(
        obj=mass_profile,
        include_critical_curves=include_critical_curves,
        include_caustics=include_caustics,
    )

    aa.plot.array(
        array=potential,
        mask=mask,
        points=positions,
        lines=lines,
        as_subplot=as_subplot,
        unit_label=unit_label,
        unit_conversion_factor=unit_conversion_factor,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        point_pointsize=position_pointsize,
        grid_pointsize=grid_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def deflections_y(
    mass_profile,
    grid,
    mask=None,
    positions=None,
    include_critical_curves=False,
    include_caustics=False,
    as_subplot=False,
    unit_label="arcsec",
    unit_conversion_factor=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Deflections (y)",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    position_pointsize=10,
    grid_pointsize=1,
    output_path=None,
    output_format="show",
    output_filename="deflections_y",
):
    """Plot the y component of the deflection angles of a mass profile, on a grid of (y,x) coordinates.

    Set *autoastro.hyper_galaxies.arrays.plotters.array_plotters* for a description of all innput parameters not described below.

    Parameters
    -----------
    mass_profile : model.profiles.mass_profiles.MassProfile
        The mass profile whose y deflecton angles are plotted.
    grid : ndarray or hyper_galaxies.arrays.grid_stacks.Grid
        The (y,x) coordinates of the grid, in an arrays of shape (total_coordinates, 2)
    """

    deflections = mass_profile.deflections_from_grid(grid=grid)
    deflections_y = grid.mapping.array_from_sub_array_1d(sub_array_1d=deflections[:, 0])

    lines = lens_plotter_util.get_critical_curves_and_caustics_from_lensing_object(
        obj=mass_profile,
        include_critical_curves=include_critical_curves,
        include_caustics=include_caustics,
    )

    aa.plot.array(
        array=deflections_y,
        mask=mask,
        points=positions,
        lines=lines,
        as_subplot=as_subplot,
        unit_label=unit_label,
        unit_conversion_factor=unit_conversion_factor,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        point_pointsize=position_pointsize,
        grid_pointsize=grid_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def deflections_x(
    mass_profile,
    grid,
    mask=None,
    positions=None,
    include_critical_curves=False,
    include_caustics=False,
    as_subplot=False,
    unit_label="arcsec",
    unit_conversion_factor=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Deflections (x)",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    position_pointsize=10,
    grid_pointsize=1,
    output_path=None,
    output_format="show",
    output_filename="deflections_x",
):
    """Plot the x component of the deflection angles of a mass profile, on a grid of (y,x) coordinates.

     Set *autoastro.hyper_galaxies.arrays.plotters.array_plotters* for a description of all innput parameters not described below.

     Parameters
     -----------
     mass_profile : model.profiles.mass_profiles.MassProfile
         The mass profile whose x deflecton angles are plotted.
     grid : ndarray or hyper_galaxies.arrays.grid_stacks.Grid
         The (y,x) coordinates of the grid, in an arrays of shape (total_coordinates, 2)
     """
    deflections = mass_profile.deflections_from_grid(grid=grid)
    deflections_x = grid.mapping.array_from_sub_array_1d(sub_array_1d=deflections[:, 1])

    lines = lens_plotter_util.get_critical_curves_and_caustics_from_lensing_object(
        obj=mass_profile,
        include_critical_curves=include_critical_curves,
        include_caustics=include_caustics,
    )

    aa.plot.array(
        array=deflections_x,
        mask=mask,
        points=positions,
        lines=lines,
        as_subplot=as_subplot,
        unit_label=unit_label,
        unit_conversion_factor=unit_conversion_factor,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        point_pointsize=position_pointsize,
        grid_pointsize=grid_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def magnification(
    mass_profile,
    grid,
    mask=None,
    positions=None,
    as_subplot=False,
    include_critical_curves=False,
    include_caustics=False,
    unit_label="arcsec",
    unit_conversion_factor=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Magnification",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    position_pointsize=10,
    grid_pointsize=1,
    output_path=None,
    output_format="show",
    output_filename="magnification",
):
    """Plot the magnification of a mass profile, on a grid of (y,x) coordinates.

    Set *autoastro.hyper_galaxies.arrays.plotters.array_plotters* for a description of all innput parameters not described below.

    Parameters
    -----------
    mass_profile : model.profiles.mass_profiles.MassProfile
        The mass profile whose magnification is plotted.
    grid : ndarray or hyper_galaxies.arrays.grid_stacks.Grid
        The (y,x) coordinates of the grid, in an arrays of shape (total_coordinates, 2)
    """
    magnification = mass_profile.magnification_from_grid(grid=grid)

    lines = lens_plotter_util.get_critical_curves_and_caustics_from_lensing_object(
        obj=mass_profile,
        include_critical_curves=include_critical_curves,
        include_caustics=include_caustics,
    )

    aa.plot.array(
        array=magnification,
        mask=mask,
        points=positions,
        lines=lines,
        as_subplot=as_subplot,
        unit_label=unit_label,
        unit_conversion_factor=unit_conversion_factor,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        point_pointsize=position_pointsize,
        grid_pointsize=grid_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )



from autoastro import dimensions as dim
from autoastro import util
from autoastro.profiles import (
    light_profiles as lp,
    mass_profiles as mp,
    light_and_mass_profiles as lmp,
)
from autoastro.galaxy.galaxy import Galaxy, HyperGalaxy, Redshift
from autoastro.galaxy.galaxy_data import GalaxyData as galaxy_data
from autoastro.galaxy import masked
from autoastro.galaxy.fit_galaxy import GalaxyFit as fit_galaxy
from autoastro.galaxy.galaxy_model import GalaxyModel
from autoastro.hyper import hyper_data
from autoastro import plotters as plot

__version__ = '0.1.4'



from __future__ import division, print_function

import math
import numpy as np
import pytest
import scipy.special

import autofit as af
import autoarray as aa
import autoastro as aast
from test_autoastro.mock import mock_cosmology


@pytest.fixture(autouse=True)
def reset_config():
    """
    Use configuration from the default path. You may want to change this to set a specific path.
    """
    af.conf.instance = af.conf.default


grid = aa.grid_irregular.manual_1d([[1.0, 1.0], [2.0, 2.0], [3.0, 3.0], [2.0, 4.0]])


class TestGaussian:
    def test__constructor_and_units(self):
        gaussian = aast.lp.EllipticalGaussian(
            centre=(1.0, 2.0), axis_ratio=0.5, phi=45.0, intensity=1.0, sigma=0.1
        )

        assert gaussian.centre == (1.0, 2.0)
        assert isinstance(gaussian.centre[0], aast.dim.Length)
        assert isinstance(gaussian.centre[1], aast.dim.Length)
        assert gaussian.centre[0].unit == "arcsec"
        assert gaussian.centre[1].unit == "arcsec"

        assert gaussian.axis_ratio == 0.5
        assert isinstance(gaussian.axis_ratio, float)

        assert gaussian.phi == 45.0
        assert isinstance(gaussian.phi, float)

        assert gaussian.intensity == 1.0
        assert isinstance(gaussian.intensity, aast.dim.Luminosity)
        assert gaussian.intensity.unit == "eps"

        assert gaussian.sigma == 0.1
        assert isinstance(gaussian.sigma, aast.dim.Length)
        assert gaussian.sigma.unit_length == "arcsec"

        gaussian = aast.lp.SphericalGaussian(
            centre=(1.0, 2.0), intensity=1.0, sigma=0.1
        )

        assert gaussian.centre == (1.0, 2.0)
        assert isinstance(gaussian.centre[0], aast.dim.Length)
        assert isinstance(gaussian.centre[1], aast.dim.Length)
        assert gaussian.centre[0].unit == "arcsec"
        assert gaussian.centre[1].unit == "arcsec"

        assert gaussian.axis_ratio == 1.0
        assert isinstance(gaussian.axis_ratio, float)

        assert gaussian.phi == 0.0
        assert isinstance(gaussian.phi, float)

        assert gaussian.intensity == 1.0
        assert isinstance(gaussian.intensity, aast.dim.Luminosity)
        assert gaussian.intensity.unit == "eps"

        assert gaussian.sigma == 0.1
        assert isinstance(gaussian.sigma, aast.dim.Length)
        assert gaussian.sigma.unit_length == "arcsec"

    def test__intensity_as_radius__correct_value(self):
        gaussian = aast.lp.EllipticalGaussian(
            centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0, intensity=1.0, sigma=1.0
        )
        assert gaussian.profile_image_from_grid_radii(grid_radii=1.0) == pytest.approx(
            0.24197, 1e-2
        )

        gaussian = aast.lp.EllipticalGaussian(
            centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0, intensity=2.0, sigma=1.0
        )
        assert gaussian.profile_image_from_grid_radii(grid_radii=1.0) == pytest.approx(
            2.0 * 0.24197, 1e-2
        )

        gaussian = aast.lp.EllipticalGaussian(
            centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0, intensity=1.0, sigma=2.0
        )
        assert gaussian.profile_image_from_grid_radii(grid_radii=1.0) == pytest.approx(
            0.1760, 1e-2
        )

        gaussian = aast.lp.EllipticalGaussian(
            centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0, intensity=1.0, sigma=2.0
        )
        assert gaussian.profile_image_from_grid_radii(grid_radii=3.0) == pytest.approx(
            0.0647, 1e-2
        )

    def test__intensity_from_grid__same_values_as_above(self):
        gaussian = aast.lp.EllipticalGaussian(
            centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0, intensity=1.0, sigma=1.0
        )
        assert gaussian.profile_image_from_grid(
            grid=aa.grid_irregular.manual_1d([[0.0, 1.0]])
        ) == pytest.approx(0.24197, 1e-2)

        gaussian = aast.lp.EllipticalGaussian(
            centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0, intensity=2.0, sigma=1.0
        )

        assert gaussian.profile_image_from_grid(
            grid=aa.grid_irregular.manual_1d([[0.0, 1.0]])
        ) == pytest.approx(2.0 * 0.24197, 1e-2)

        gaussian = aast.lp.EllipticalGaussian(
            centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0, intensity=1.0, sigma=2.0
        )

        assert gaussian.profile_image_from_grid(
            grid=aa.grid_irregular.manual_1d([[0.0, 1.0]])
        ) == pytest.approx(0.1760, 1e-2)

        gaussian = aast.lp.EllipticalGaussian(
            centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0, intensity=1.0, sigma=2.0
        )

        assert gaussian.profile_image_from_grid(
            grid=aa.grid_irregular.manual_1d([[0.0, 3.0]])
        ) == pytest.approx(0.0647, 1e-2)

    def test__intensity_from_grid__change_geometry(self):
        gaussian = aast.lp.EllipticalGaussian(
            centre=(1.0, 1.0), axis_ratio=1.0, phi=0.0, intensity=1.0, sigma=1.0
        )
        assert gaussian.profile_image_from_grid(
            grid=aa.grid_irregular.manual_1d([[1.0, 0.0]])
        ) == pytest.approx(0.24197, 1e-2)

        gaussian = aast.lp.EllipticalGaussian(
            centre=(0.0, 0.0), axis_ratio=0.5, phi=0.0, intensity=1.0, sigma=1.0
        )
        assert gaussian.profile_image_from_grid(
            grid=aa.grid_irregular.manual_1d([[1.0, 0.0]])
        ) == pytest.approx(0.05399, 1e-2)

        gaussian_0 = aast.lp.EllipticalGaussian(
            centre=(-3.0, -0.0), axis_ratio=0.5, phi=0.0, intensity=1.0, sigma=1.0
        )

        gaussian_1 = aast.lp.EllipticalGaussian(
            centre=(3.0, 0.0), axis_ratio=0.5, phi=0.0, intensity=1.0, sigma=1.0
        )

        assert gaussian_0.profile_image_from_grid(
            grid=aa.grid_irregular.manual_1d([[0.0, 0.0], [0.0, 1.0], [0.0, -1.0]])
        ) == pytest.approx(
            gaussian_1.profile_image_from_grid(
                grid=aa.grid_irregular.manual_1d([[0.0, 0.0], [0.0, 1.0], [0.0, -1.0]])
            ),
            1e-4,
        )

        gaussian_0 = aast.lp.EllipticalGaussian(
            centre=(0.0, 0.0), axis_ratio=0.5, phi=180.0, intensity=1.0, sigma=1.0
        )

        gaussian_1 = aast.lp.EllipticalGaussian(
            centre=(0.0, 0.0), axis_ratio=0.5, phi=0.0, intensity=1.0, sigma=1.0
        )

        assert gaussian_0.profile_image_from_grid(
            grid=aa.grid_irregular.manual_1d([[0.0, 0.0], [0.0, 1.0], [0.0, -1.0]])
        ) == pytest.approx(
            gaussian_1.profile_image_from_grid(
                grid=aa.grid_irregular.manual_1d([[0.0, 0.0], [0.0, -1.0], [0.0, 1.0]])
            ),
            1e-4,
        )

    def test__spherical_and_elliptical_match(self):
        elliptical = aast.lp.EllipticalGaussian(
            axis_ratio=1.0, phi=0.0, intensity=3.0, sigma=2.0
        )
        spherical = aast.lp.SphericalGaussian(intensity=3.0, sigma=2.0)

        assert (
            elliptical.profile_image_from_grid(grid=grid)
            == spherical.profile_image_from_grid(grid=grid)
        ).all()

    def test__output_image_is_autoarray(self):
        grid = aa.grid.uniform(shape_2d=(2, 2), pixel_scales=1.0, sub_size=1)

        gaussian = aast.lp.EllipticalGaussian()

        image = gaussian.profile_image_from_grid(grid=grid)

        assert image.shape_2d == (2, 2)

        gaussian = aast.lp.SphericalGaussian()

        image = gaussian.profile_image_from_grid(grid=grid)

        assert image.shape_2d == (2, 2)


class TestSersic:
    def test__constructor_and_units(self):
        sersic = aast.lp.EllipticalSersic(
            centre=(1.0, 2.0),
            axis_ratio=0.5,
            phi=45.0,
            intensity=1.0,
            effective_radius=0.6,
            sersic_index=4.0,
        )

        assert sersic.centre == (1.0, 2.0)
        assert isinstance(sersic.centre[0], aast.dim.Length)
        assert isinstance(sersic.centre[1], aast.dim.Length)
        assert sersic.centre[0].unit == "arcsec"
        assert sersic.centre[1].unit == "arcsec"

        assert sersic.axis_ratio == 0.5
        assert isinstance(sersic.axis_ratio, float)

        assert sersic.phi == 45.0
        assert isinstance(sersic.phi, float)

        assert sersic.intensity == 1.0
        assert isinstance(sersic.intensity, aast.dim.Luminosity)
        assert sersic.intensity.unit == "eps"

        assert sersic.effective_radius == 0.6
        assert isinstance(sersic.effective_radius, aast.dim.Length)
        assert sersic.effective_radius.unit_length == "arcsec"

        assert sersic.sersic_index == 4.0
        assert isinstance(sersic.sersic_index, float)

        assert sersic.sersic_constant == pytest.approx(7.66925, 1e-3)
        assert sersic.elliptical_effective_radius == 0.6 / np.sqrt(0.5)

        sersic = aast.lp.SphericalSersic(
            centre=(1.0, 2.0), intensity=1.0, effective_radius=0.6, sersic_index=4.0
        )

        assert sersic.centre == (1.0, 2.0)
        assert isinstance(sersic.centre[0], aast.dim.Length)
        assert isinstance(sersic.centre[1], aast.dim.Length)
        assert sersic.centre[0].unit == "arcsec"
        assert sersic.centre[1].unit == "arcsec"

        assert sersic.axis_ratio == 1.0
        assert isinstance(sersic.axis_ratio, float)

        assert sersic.phi == 0.0
        assert isinstance(sersic.phi, float)

        assert sersic.intensity == 1.0
        assert isinstance(sersic.intensity, aast.dim.Luminosity)
        assert sersic.intensity.unit == "eps"

        assert sersic.effective_radius == 0.6
        assert isinstance(sersic.effective_radius, aast.dim.Length)
        assert sersic.effective_radius.unit_length == "arcsec"

        assert sersic.sersic_index == 4.0
        assert isinstance(sersic.sersic_index, float)

        assert sersic.sersic_constant == pytest.approx(7.66925, 1e-3)
        assert sersic.elliptical_effective_radius == 0.6

    def test__intensity_at_radius__correct_value(self):
        sersic = aast.lp.EllipticalSersic(
            axis_ratio=1.0,
            phi=0.0,
            intensity=1.0,
            effective_radius=0.6,
            sersic_index=4.0,
        )
        assert sersic.profile_image_from_grid_radii(grid_radii=1.0) == pytest.approx(
            0.351797, 1e-3
        )

        sersic = aast.lp.EllipticalSersic(
            axis_ratio=1.0,
            phi=0.0,
            intensity=3.0,
            effective_radius=2.0,
            sersic_index=2.0,
        )
        # 3.0 * exp(-3.67206544592 * (1,5/2.0) ** (1.0 / 2.0)) - 1) = 0.351797
        assert sersic.profile_image_from_grid_radii(grid_radii=1.5) == pytest.approx(
            4.90657319276, 1e-3
        )

    def test__intensity_from_grid__correct_values(self):
        sersic = aast.lp.EllipticalSersic(
            axis_ratio=0.5,
            phi=0.0,
            intensity=3.0,
            effective_radius=2.0,
            sersic_index=2.0,
        )
        assert sersic.profile_image_from_grid(
            grid=aa.grid_irregular.manual_1d([[1.0, 0.0]])
        ) == pytest.approx(5.38066670129, 1e-3)

    def test__intensity_from_grid__change_geometry(self):
        sersic_0 = aast.lp.EllipticalSersic(
            axis_ratio=0.5,
            phi=0.0,
            intensity=3.0,
            effective_radius=2.0,
            sersic_index=2.0,
        )

        sersic_1 = aast.lp.EllipticalSersic(
            axis_ratio=0.5,
            phi=90.0,
            intensity=3.0,
            effective_radius=2.0,
            sersic_index=2.0,
        )

        assert sersic_0.profile_image_from_grid(
            grid=aa.grid_irregular.manual_1d([[0.0, 1.0]])
        ) == sersic_1.profile_image_from_grid(
            grid=aa.grid_irregular.manual_1d([[1.0, 0.0]])
        )

    def test__spherical_and_elliptical_match(self):
        elliptical = aast.lp.EllipticalSersic(
            axis_ratio=1.0,
            phi=0.0,
            intensity=3.0,
            effective_radius=2.0,
            sersic_index=2.0,
        )

        spherical = aast.lp.SphericalSersic(
            intensity=3.0, effective_radius=2.0, sersic_index=2.0
        )

        assert (
            elliptical.profile_image_from_grid(grid=grid)
            == spherical.profile_image_from_grid(grid=grid)
        ).all()

    def test__summarize_in_units(self):
        sersic = aast.lp.SphericalSersic(
            intensity=3.0, effective_radius=2.0, sersic_index=2.0
        )

        summary_text = sersic.summarize_in_units(
            radii=[aast.dim.Length(10.0), aast.dim.Length(500.0)],
            prefix="sersic_",
            unit_length="arcsec",
            unit_luminosity="eps",
            whitespace=50,
        )

        i = 0

        assert summary_text[i] == "Light Profile = SphericalSersic\n"
        i += 1
        assert (
            summary_text[i]
            == "sersic_luminosity_within_10.00_arcsec             1.8854e+02 eps"
        )
        i += 1
        assert (
            summary_text[i]
            == "sersic_luminosity_within_500.00_arcsec            1.9573e+02 eps"
        )
        i += 1

    def test__output_image_is_autoarray(self):
        grid = aa.grid.uniform(shape_2d=(2, 2), pixel_scales=1.0, sub_size=1)

        sersic = aast.lp.EllipticalSersic()

        image = sersic.profile_image_from_grid(grid=grid)

        assert image.shape_2d == (2, 2)

        sersic = aast.lp.SphericalSersic()

        image = sersic.profile_image_from_grid(grid=grid)

        assert image.shape_2d == (2, 2)


class TestExponential:
    def test__constructor_and_units(self):
        exponential = aast.lp.EllipticalExponential(
            centre=(1.0, 2.0),
            axis_ratio=0.5,
            phi=45.0,
            intensity=1.0,
            effective_radius=0.6,
        )

        assert exponential.centre == (1.0, 2.0)
        assert isinstance(exponential.centre[0], aast.dim.Length)
        assert isinstance(exponential.centre[1], aast.dim.Length)
        assert exponential.centre[0].unit == "arcsec"
        assert exponential.centre[1].unit == "arcsec"

        assert exponential.axis_ratio == 0.5
        assert isinstance(exponential.axis_ratio, float)

        assert exponential.phi == 45.0
        assert isinstance(exponential.phi, float)

        assert exponential.intensity == 1.0
        assert isinstance(exponential.intensity, aast.dim.Luminosity)
        assert exponential.intensity.unit == "eps"

        assert exponential.effective_radius == 0.6
        assert isinstance(exponential.effective_radius, aast.dim.Length)
        assert exponential.effective_radius.unit_length == "arcsec"

        assert exponential.sersic_index == 1.0
        assert isinstance(exponential.sersic_index, float)

        assert exponential.sersic_constant == pytest.approx(1.67838, 1e-3)
        assert exponential.elliptical_effective_radius == 0.6 / np.sqrt(0.5)

        exponential = aast.lp.SphericalExponential(
            centre=(1.0, 2.0), intensity=1.0, effective_radius=0.6
        )

        assert exponential.centre == (1.0, 2.0)
        assert isinstance(exponential.centre[0], aast.dim.Length)
        assert isinstance(exponential.centre[1], aast.dim.Length)
        assert exponential.centre[0].unit == "arcsec"
        assert exponential.centre[1].unit == "arcsec"

        assert exponential.axis_ratio == 1.0
        assert isinstance(exponential.axis_ratio, float)

        assert exponential.phi == 0.0
        assert isinstance(exponential.phi, float)

        assert exponential.intensity == 1.0
        assert isinstance(exponential.intensity, aast.dim.Luminosity)
        assert exponential.intensity.unit == "eps"

        assert exponential.effective_radius == 0.6
        assert isinstance(exponential.effective_radius, aast.dim.Length)
        assert exponential.effective_radius.unit_length == "arcsec"

        assert exponential.sersic_index == 1.0
        assert isinstance(exponential.sersic_index, float)

        assert exponential.sersic_constant == pytest.approx(1.67838, 1e-3)
        assert exponential.elliptical_effective_radius == 0.6

    def test__intensity_at_radius__correct_value(self):
        exponential = aast.lp.EllipticalExponential(
            axis_ratio=1.0, phi=0.0, intensity=1.0, effective_radius=0.6
        )
        assert exponential.profile_image_from_grid_radii(
            grid_radii=1.0
        ) == pytest.approx(0.3266, 1e-3)

        exponential = aast.lp.EllipticalExponential(
            axis_ratio=1.0, phi=0.0, intensity=3.0, effective_radius=2.0
        )
        assert exponential.profile_image_from_grid_radii(
            grid_radii=1.5
        ) == pytest.approx(4.5640, 1e-3)

    def test__intensity_from_grid__correct_values(self):
        exponential = aast.lp.EllipticalExponential(
            axis_ratio=0.5, phi=0.0, intensity=3.0, effective_radius=2.0
        )
        assert exponential.profile_image_from_grid(
            grid=aa.grid_irregular.manual_1d([[1.0, 0.0]])
        ) == pytest.approx(4.9047, 1e-3)

        exponential = aast.lp.EllipticalExponential(
            axis_ratio=0.5, phi=90.0, intensity=2.0, effective_radius=3.0
        )
        assert exponential.profile_image_from_grid(
            grid=aa.grid_irregular.manual_1d([[0.0, 1.0]])
        ) == pytest.approx(4.8566, 1e-3)

        exponential = aast.lp.EllipticalExponential(
            axis_ratio=0.5, phi=90.0, intensity=4.0, effective_radius=3.0
        )
        assert exponential.profile_image_from_grid(
            grid=aa.grid_irregular.manual_1d([[0.0, 1.0]])
        ) == pytest.approx(2.0 * 4.8566, 1e-3)

    def test__intensity_from_grid__change_geometry(self):
        exponential_0 = aast.lp.EllipticalExponential(
            axis_ratio=0.5, phi=0.0, intensity=3.0, effective_radius=2.0
        )

        exponential_1 = aast.lp.EllipticalExponential(
            axis_ratio=0.5, phi=90.0, intensity=3.0, effective_radius=2.0
        )

        assert exponential_0.profile_image_from_grid(
            grid=aa.grid_irregular.manual_1d([[0.0, 1.0]])
        ) == exponential_1.profile_image_from_grid(
            grid=aa.grid_irregular.manual_1d([[1.0, 0.0]])
        )

    def test__spherical_and_elliptical_match(self):
        elliptical = aast.lp.EllipticalExponential(
            axis_ratio=1.0, phi=0.0, intensity=3.0, effective_radius=2.0
        )

        spherical = aast.lp.SphericalExponential(intensity=3.0, effective_radius=2.0)

        assert (
            elliptical.profile_image_from_grid(grid=grid)
            == spherical.profile_image_from_grid(grid=grid)
        ).all()

    def test__output_image_is_autoarray(self):
        grid = aa.grid.uniform(shape_2d=(2, 2), pixel_scales=1.0, sub_size=1)

        exponential = aast.lp.EllipticalExponential()

        image = exponential.profile_image_from_grid(grid=grid)

        assert image.shape_2d == (2, 2)

        exponential = aast.lp.SphericalExponential()

        image = exponential.profile_image_from_grid(grid=grid)

        assert image.shape_2d == (2, 2)


class TestDevVaucouleurs:
    def test__constructor_and_units(self):
        dev_vaucouleurs = aast.lp.EllipticalDevVaucouleurs(
            centre=(1.0, 2.0),
            axis_ratio=0.5,
            phi=45.0,
            intensity=1.0,
            effective_radius=0.6,
        )

        assert dev_vaucouleurs.centre == (1.0, 2.0)
        assert isinstance(dev_vaucouleurs.centre[0], aast.dim.Length)
        assert isinstance(dev_vaucouleurs.centre[1], aast.dim.Length)
        assert dev_vaucouleurs.centre[0].unit == "arcsec"
        assert dev_vaucouleurs.centre[1].unit == "arcsec"

        assert dev_vaucouleurs.axis_ratio == 0.5
        assert isinstance(dev_vaucouleurs.axis_ratio, float)

        assert dev_vaucouleurs.phi == 45.0
        assert isinstance(dev_vaucouleurs.phi, float)

        assert dev_vaucouleurs.intensity == 1.0
        assert isinstance(dev_vaucouleurs.intensity, aast.dim.Luminosity)
        assert dev_vaucouleurs.intensity.unit == "eps"

        assert dev_vaucouleurs.effective_radius == 0.6
        assert isinstance(dev_vaucouleurs.effective_radius, aast.dim.Length)
        assert dev_vaucouleurs.effective_radius.unit_length == "arcsec"

        assert dev_vaucouleurs.sersic_index == 4.0
        assert isinstance(dev_vaucouleurs.sersic_index, float)

        assert dev_vaucouleurs.sersic_constant == pytest.approx(7.66924, 1e-3)
        assert dev_vaucouleurs.elliptical_effective_radius == 0.6 / np.sqrt(0.5)

        dev_vaucouleurs = aast.lp.SphericalDevVaucouleurs(
            centre=(1.0, 2.0), intensity=1.0, effective_radius=0.6
        )

        assert dev_vaucouleurs.centre == (1.0, 2.0)
        assert isinstance(dev_vaucouleurs.centre[0], aast.dim.Length)
        assert isinstance(dev_vaucouleurs.centre[1], aast.dim.Length)
        assert dev_vaucouleurs.centre[0].unit == "arcsec"
        assert dev_vaucouleurs.centre[1].unit == "arcsec"

        assert dev_vaucouleurs.axis_ratio == 1.0
        assert isinstance(dev_vaucouleurs.axis_ratio, float)

        assert dev_vaucouleurs.phi == 0.0
        assert isinstance(dev_vaucouleurs.phi, float)

        assert dev_vaucouleurs.intensity == 1.0
        assert isinstance(dev_vaucouleurs.intensity, aast.dim.Luminosity)
        assert dev_vaucouleurs.intensity.unit == "eps"

        assert dev_vaucouleurs.effective_radius == 0.6
        assert isinstance(dev_vaucouleurs.effective_radius, aast.dim.Length)
        assert dev_vaucouleurs.effective_radius.unit_length == "arcsec"

        assert dev_vaucouleurs.sersic_index == 4.0
        assert isinstance(dev_vaucouleurs.sersic_index, float)

        assert dev_vaucouleurs.sersic_constant == pytest.approx(7.66924, 1e-3)
        assert dev_vaucouleurs.elliptical_effective_radius == 0.6

    def test__intensity_at_radius__correct_value(self):
        dev_vaucouleurs = aast.lp.EllipticalDevVaucouleurs(
            axis_ratio=1.0, phi=0.0, intensity=1.0, effective_radius=0.6
        )
        assert dev_vaucouleurs.profile_image_from_grid_radii(
            grid_radii=1.0
        ) == pytest.approx(0.3518, 1e-3)

        dev_vaucouleurs = aast.lp.EllipticalDevVaucouleurs(
            axis_ratio=1.0, phi=0.0, intensity=3.0, effective_radius=2.0
        )
        assert dev_vaucouleurs.profile_image_from_grid_radii(
            grid_radii=1.5
        ) == pytest.approx(5.1081, 1e-3)

    def test__intensity_from_grid__correct_values(self):
        dev_vaucouleurs = aast.lp.EllipticalDevVaucouleurs(
            axis_ratio=0.5, phi=0.0, intensity=3.0, effective_radius=2.0
        )
        assert dev_vaucouleurs.profile_image_from_grid(
            grid=aa.grid_irregular.manual_1d([[1.0, 0.0]])
        ) == pytest.approx(5.6697, 1e-3)

        dev_vaucouleurs = aast.lp.EllipticalDevVaucouleurs(
            axis_ratio=0.5, phi=90.0, intensity=2.0, effective_radius=3.0
        )

        assert dev_vaucouleurs.profile_image_from_grid(
            grid=aa.grid_irregular.manual_1d([[0.0, 1.0]])
        ) == pytest.approx(7.4455, 1e-3)

        dev_vaucouleurs = aast.lp.EllipticalDevVaucouleurs(
            axis_ratio=0.5, phi=90.0, intensity=4.0, effective_radius=3.0
        )
        assert dev_vaucouleurs.profile_image_from_grid(
            grid=aa.grid_irregular.manual_1d([[0.0, 1.0]])
        ) == pytest.approx(2.0 * 7.4455, 1e-3)

    def test__intensity_from_grid__change_geometry(self):
        dev_vaucouleurs_0 = aast.lp.EllipticalDevVaucouleurs(
            axis_ratio=0.5, phi=0.0, intensity=3.0, effective_radius=2.0
        )

        dev_vaucouleurs_1 = aast.lp.EllipticalDevVaucouleurs(
            axis_ratio=0.5, phi=90.0, intensity=3.0, effective_radius=2.0
        )

        assert dev_vaucouleurs_0.profile_image_from_grid(
            grid=aa.grid_irregular.manual_1d([[0.0, 1.0]])
        ) == dev_vaucouleurs_1.profile_image_from_grid(
            grid=aa.grid_irregular.manual_1d([[1.0, 0.0]])
        )

    def test__spherical_and_elliptical_match(self):
        elliptical = aast.lp.EllipticalDevVaucouleurs(
            axis_ratio=1.0, phi=0.0, intensity=3.0, effective_radius=2.0
        )

        spherical = aast.lp.SphericalDevVaucouleurs(intensity=3.0, effective_radius=2.0)

        assert (
            elliptical.profile_image_from_grid(grid=grid)
            == spherical.profile_image_from_grid(grid=grid)
        ).all()

    def test__output_image_is_autoarray(self):
        grid = aa.grid.uniform(shape_2d=(2, 2), pixel_scales=1.0, sub_size=1)

        dev_vaucouleurs = aast.lp.EllipticalDevVaucouleurs()

        image = dev_vaucouleurs.profile_image_from_grid(grid=grid)

        assert image.shape_2d == (2, 2)

        dev_vaucouleurs = aast.lp.SphericalDevVaucouleurs()

        image = dev_vaucouleurs.profile_image_from_grid(grid=grid)

        assert image.shape_2d == (2, 2)


class TestCoreSersic(object):
    def test__constructor_and_units(self):
        core_sersic = aast.lp.EllipticalCoreSersic(
            centre=(1.0, 2.0),
            axis_ratio=0.5,
            phi=45.0,
            intensity=1.0,
            effective_radius=0.6,
            sersic_index=4.0,
            radius_break=0.01,
            intensity_break=0.1,
            gamma=1.0,
            alpha=2.0,
        )

        assert core_sersic.centre == (1.0, 2.0)
        assert isinstance(core_sersic.centre[0], aast.dim.Length)
        assert isinstance(core_sersic.centre[1], aast.dim.Length)
        assert core_sersic.centre[0].unit == "arcsec"
        assert core_sersic.centre[1].unit == "arcsec"

        assert core_sersic.axis_ratio == 0.5
        assert isinstance(core_sersic.axis_ratio, float)

        assert core_sersic.phi == 45.0
        assert isinstance(core_sersic.phi, float)

        assert core_sersic.intensity == 1.0
        assert isinstance(core_sersic.intensity, aast.dim.Luminosity)
        assert core_sersic.intensity.unit == "eps"

        assert core_sersic.effective_radius == 0.6
        assert isinstance(core_sersic.effective_radius, aast.dim.Length)
        assert core_sersic.effective_radius.unit_length == "arcsec"

        assert core_sersic.sersic_index == 4.0
        assert isinstance(core_sersic.sersic_index, float)

        assert core_sersic.radius_break == 0.01
        assert isinstance(core_sersic.radius_break, aast.dim.Length)
        assert core_sersic.radius_break.unit_length == "arcsec"

        assert core_sersic.intensity_break == 0.1
        assert isinstance(core_sersic.intensity_break, aast.dim.Luminosity)
        assert core_sersic.intensity_break.unit == "eps"

        assert core_sersic.gamma == 1.0
        assert isinstance(core_sersic.gamma, float)

        assert core_sersic.alpha == 2.0
        assert isinstance(core_sersic.alpha, float)

        assert core_sersic.sersic_constant == pytest.approx(7.66925, 1e-3)
        assert core_sersic.elliptical_effective_radius == 0.6 / np.sqrt(0.5)

        core_sersic = aast.lp.SphericalCoreSersic(
            centre=(1.0, 2.0),
            intensity=1.0,
            effective_radius=0.6,
            sersic_index=4.0,
            radius_break=0.01,
            intensity_break=0.1,
            gamma=1.0,
            alpha=2.0,
        )

        assert core_sersic.centre == (1.0, 2.0)
        assert isinstance(core_sersic.centre[0], aast.dim.Length)
        assert isinstance(core_sersic.centre[1], aast.dim.Length)
        assert core_sersic.centre[0].unit == "arcsec"
        assert core_sersic.centre[1].unit == "arcsec"

        assert core_sersic.axis_ratio == 1.0
        assert isinstance(core_sersic.axis_ratio, float)

        assert core_sersic.phi == 0.0
        assert isinstance(core_sersic.phi, float)

        assert core_sersic.intensity == 1.0
        assert isinstance(core_sersic.intensity, aast.dim.Luminosity)
        assert core_sersic.intensity.unit == "eps"

        assert core_sersic.effective_radius == 0.6
        assert isinstance(core_sersic.effective_radius, aast.dim.Length)
        assert core_sersic.effective_radius.unit_length == "arcsec"

        assert core_sersic.sersic_index == 4.0
        assert isinstance(core_sersic.sersic_index, float)

        assert core_sersic.radius_break == 0.01
        assert isinstance(core_sersic.radius_break, aast.dim.Length)
        assert core_sersic.radius_break.unit_length == "arcsec"

        assert core_sersic.intensity_break == 0.1
        assert isinstance(core_sersic.intensity_break, aast.dim.Luminosity)
        assert core_sersic.intensity_break.unit == "eps"

        assert core_sersic.gamma == 1.0
        assert isinstance(core_sersic.gamma, float)

        assert core_sersic.alpha == 2.0
        assert isinstance(core_sersic.alpha, float)

        assert core_sersic.sersic_constant == pytest.approx(7.66925, 1e-3)
        assert core_sersic.elliptical_effective_radius == 0.6

    def test__intensity_at_radius__correct_value(self):
        core_sersic = aast.lp.EllipticalCoreSersic(
            axis_ratio=0.5,
            phi=0.0,
            intensity=1.0,
            effective_radius=5.0,
            sersic_index=4.0,
            radius_break=0.01,
            intensity_break=0.1,
            gamma=1.0,
            alpha=1.0,
        )
        assert core_sersic.profile_image_from_grid_radii(0.01) == 0.1

    def test__spherical_and_elliptical_match(self):
        elliptical = aast.lp.EllipticalCoreSersic(
            axis_ratio=1.0,
            phi=0.0,
            intensity=1.0,
            effective_radius=5.0,
            sersic_index=4.0,
            radius_break=0.01,
            intensity_break=0.1,
            gamma=1.0,
            alpha=1.0,
        )

        spherical = aast.lp.SphericalCoreSersic(
            intensity=1.0,
            effective_radius=5.0,
            sersic_index=4.0,
            radius_break=0.01,
            intensity_break=0.1,
            gamma=1.0,
            alpha=1.0,
        )

        assert (
            elliptical.profile_image_from_grid(grid=grid)
            == spherical.profile_image_from_grid(grid=grid)
        ).all()

    def test__output_image_is_autoarray(self):
        grid = aa.grid.uniform(shape_2d=(2, 2), pixel_scales=1.0, sub_size=1)

        core_sersic = aast.lp.EllipticalCoreSersic()

        image = core_sersic.profile_image_from_grid(grid=grid)

        assert image.shape_2d == (2, 2)

        core_sersic = aast.lp.SphericalCoreSersic()

        image = core_sersic.profile_image_from_grid(grid=grid)

        assert image.shape_2d == (2, 2)


class TestBlurredProfileImages(object):
    def test__blurred_image_from_grid_and_psf(
        self, sub_grid_7x7, blurring_grid_7x7, psf_3x3, convolver_7x7
    ):

        light_profile = aast.lp.EllipticalSersic(intensity=1.0)

        image = light_profile.profile_image_from_grid(grid=sub_grid_7x7)

        blurring_image = light_profile.profile_image_from_grid(grid=blurring_grid_7x7)

        blurred_image = convolver_7x7.convolved_image_from_image_and_blurring_image(
            image=image.in_1d_binned, blurring_image=blurring_image.in_1d_binned
        )

        light_profile_blurred_image = light_profile.blurred_profile_image_from_grid_and_psf(
            grid=sub_grid_7x7, blurring_grid=blurring_grid_7x7, psf=psf_3x3
        )

        assert blurred_image.in_1d == pytest.approx(
            light_profile_blurred_image.in_1d, 1.0e-4
        )
        assert blurred_image.in_2d == pytest.approx(
            light_profile_blurred_image.in_2d, 1.0e-4
        )

    def test__blurred_image_from_grid_and_convolver(
        self, sub_grid_7x7, blurring_grid_7x7, convolver_7x7
    ):

        light_profile = aast.lp.EllipticalSersic(intensity=1.0)

        image = light_profile.profile_image_from_grid(grid=sub_grid_7x7)

        blurring_image = light_profile.profile_image_from_grid(grid=blurring_grid_7x7)

        blurred_image = convolver_7x7.convolved_image_from_image_and_blurring_image(
            image=image.in_1d_binned, blurring_image=blurring_image.in_1d_binned
        )

        light_profile_blurred_image = light_profile.blurred_profile_image_from_grid_and_convolver(
            grid=sub_grid_7x7, convolver=convolver_7x7, blurring_grid=blurring_grid_7x7
        )

        assert blurred_image.in_1d == pytest.approx(
            light_profile_blurred_image.in_1d, 1.0e-4
        )
        assert blurred_image.in_2d == pytest.approx(
            light_profile_blurred_image.in_2d, 1.0e-4
        )


class TestVisibilities(object):
    def test__visibilities_from_grid_and_transformer(
        self, grid_7x7, sub_grid_7x7, transformer_7x7_7
    ):
        light_profile = aast.lp.EllipticalSersic(intensity=1.0)

        image = light_profile.profile_image_from_grid(grid=grid_7x7)

        visibilities = transformer_7x7_7.visibilities_from_image(
            image=image.in_1d_binned
        )

        light_profile_visibilities = light_profile.profile_visibilities_from_grid_and_transformer(
            grid=grid_7x7, transformer=transformer_7x7_7
        )

        assert visibilities == pytest.approx(light_profile_visibilities, 1.0e-4)


def luminosity_from_radius_and_profile(radius, profile):
    x = profile.sersic_constant * (
        (radius / profile.effective_radius) ** (1.0 / profile.sersic_index)
    )

    return (
        profile.intensity
        * profile.effective_radius ** 2
        * 2
        * math.pi
        * profile.sersic_index
        * (
            (math.e ** profile.sersic_constant)
            / (profile.sersic_constant ** (2 * profile.sersic_index))
        )
        * scipy.special.gamma(2 * profile.sersic_index)
        * scipy.special.gammainc(2 * profile.sersic_index, x)
    )


class TestLuminosityWithinCircle(object):
    def test__luminosity_in_eps__spherical_sersic_index_2__compare_to_analytic(self):
        sersic = aast.lp.SphericalSersic(
            intensity=3.0, effective_radius=2.0, sersic_index=2.0
        )

        radius = aast.dim.Length(0.5, "arcsec")

        luminosity_analytic = luminosity_from_radius_and_profile(
            radius=radius, profile=sersic
        )

        luminosity_integral = sersic.luminosity_within_circle_in_units(
            radius=0.5, unit_luminosity="eps"
        )

        assert luminosity_analytic == pytest.approx(luminosity_integral, 1e-3)

    def test__luminosity_in_eps__spherical_sersic_2__compare_to_grid(self):
        sersic = aast.lp.SphericalSersic(
            intensity=3.0, effective_radius=2.0, sersic_index=2.0
        )

        radius = aast.dim.Length(1.0, "arcsec")

        luminosity_grid = luminosity_from_radius_and_profile(
            radius=radius, profile=sersic
        )

        luminosity_integral = sersic.luminosity_within_circle_in_units(
            radius=radius, unit_luminosity="eps"
        )

        assert luminosity_grid == pytest.approx(luminosity_integral, 0.02)

    def test__luminosity_units_conversions__uses_exposure_time(self):
        sersic_eps = aast.lp.SphericalSersic(
            intensity=aast.dim.Luminosity(3.0, "eps"),
            effective_radius=2.0,
            sersic_index=1.0,
        )

        radius = aast.dim.Length(0.5, "arcsec")

        luminosity_analytic = luminosity_from_radius_and_profile(
            radius=radius, profile=sersic_eps
        )

        luminosity_integral = sersic_eps.luminosity_within_circle_in_units(
            radius=radius, unit_luminosity="eps", exposure_time=3.0
        )

        # eps -> eps

        assert luminosity_analytic == pytest.approx(luminosity_integral, 1e-3)

        # eps -> counts

        luminosity_integral = sersic_eps.luminosity_within_circle_in_units(
            radius=radius, unit_luminosity="counts", exposure_time=3.0
        )

        assert 3.0 * luminosity_analytic == pytest.approx(luminosity_integral, 1e-3)

        sersic_counts = aast.lp.SphericalSersic(
            intensity=aast.dim.Luminosity(3.0, "counts"),
            effective_radius=2.0,
            sersic_index=1.0,
        )

        radius = aast.dim.Length(0.5, "arcsec")

        luminosity_analytic = luminosity_from_radius_and_profile(
            radius=radius, profile=sersic_counts
        )
        luminosity_integral = sersic_counts.luminosity_within_circle_in_units(
            radius=radius, unit_luminosity="eps", exposure_time=3.0
        )

        # counts -> eps

        assert luminosity_analytic / 3.0 == pytest.approx(luminosity_integral, 1e-3)

        luminosity_integral = sersic_counts.luminosity_within_circle_in_units(
            radius=radius, unit_luminosity="counts", exposure_time=3.0
        )

        # counts -> counts

        assert luminosity_analytic == pytest.approx(luminosity_integral, 1e-3)

    def test__radius_units_conversions__light_profile_updates_units_and_computes_correct_luminosity(
        self
    ):
        cosmology = mock_cosmology.MockCosmology(arcsec_per_kpc=0.5, kpc_per_arcsec=2.0)

        sersic_arcsec = aast.lp.SphericalSersic(
            centre=(aast.dim.Length(0.0, "arcsec"), aast.dim.Length(0.0, "arcsec")),
            intensity=aast.dim.Luminosity(3.0, "eps"),
            effective_radius=aast.dim.Length(2.0, "arcsec"),
            sersic_index=1.0,
        )

        sersic_kpc = aast.lp.SphericalSersic(
            centre=(aast.dim.Length(0.0, "kpc"), aast.dim.Length(0.0, "kpc")),
            intensity=aast.dim.Luminosity(3.0, "eps"),
            effective_radius=aast.dim.Length(4.0, "kpc"),
            sersic_index=1.0,
        )

        radius = aast.dim.Length(0.5, "arcsec")

        luminosity_analytic = luminosity_from_radius_and_profile(
            radius=radius, profile=sersic_arcsec
        )

        # arcsec -> arcsec

        luminosity = sersic_arcsec.luminosity_within_circle_in_units(radius=radius)

        assert luminosity_analytic == pytest.approx(luminosity, 1e-3)

        # kpc -> arcsec

        luminosity_analytic = luminosity_from_radius_and_profile(
            radius=1.0, profile=sersic_kpc
        )

        luminosity = sersic_kpc.luminosity_within_circle_in_units(
            radius=radius, redshift_object=0.5, cosmology=cosmology
        )

        assert luminosity_analytic == pytest.approx(luminosity, 1e-3)

        radius = aast.dim.Length(0.5, "kpc")

        luminosity_analytic = luminosity_from_radius_and_profile(
            radius=radius, profile=sersic_kpc
        )

        # kpc -> kpc

        luminosity = sersic_kpc.luminosity_within_circle_in_units(radius=radius)

        assert luminosity_analytic == pytest.approx(luminosity, 1e-3)

        # kpc -> arcsec

        luminosity_analytic = luminosity_from_radius_and_profile(
            radius=0.25, profile=sersic_arcsec
        )

        luminosity = sersic_arcsec.luminosity_within_circle_in_units(
            radius=radius, redshift_object=0.5, cosmology=cosmology
        )

        assert luminosity_analytic == pytest.approx(luminosity, 1e-3)

        radius = aast.dim.Length(2.0, "arcsec")
        luminosity_arcsec = sersic_arcsec.luminosity_within_circle_in_units(
            radius=radius, redshift_object=0.5, unit_mass="angular", cosmology=cosmology
        )
        radius = aast.dim.Length(4.0, "kpc")
        luminosity_kpc = sersic_arcsec.luminosity_within_circle_in_units(
            radius=radius, redshift_object=0.5, unit_mass="angular", cosmology=cosmology
        )
        assert luminosity_arcsec == luminosity_kpc


class TestGrids(object):
    def test__grid_to_eccentric_radius(self):
        elliptical = aast.lp.EllipticalSersic(axis_ratio=0.5, phi=0.0)

        assert elliptical.grid_to_eccentric_radii(
            aa.grid_irregular.manual_1d([[1, 1]])
        ) == pytest.approx(
            elliptical.grid_to_eccentric_radii(aa.grid_irregular.manual_1d([[-1, -1]])),
            1e-10,
        )

    def test__intensity_from_grid(self):
        elliptical = aast.lp.EllipticalSersic(axis_ratio=0.5, phi=0.0)

        assert elliptical.profile_image_from_grid(
            aa.grid_irregular.manual_1d([[1, 1]])
        ) == pytest.approx(
            elliptical.profile_image_from_grid(aa.grid_irregular.manual_1d([[-1, -1]])),
            1e-4,
        )



import os

import pytest
from matplotlib import pyplot


class PlotPatch(object):
    def __init__(self):
        self.paths = []

    def __call__(self, path, *args, **kwargs):
        self.paths.append(path)


@pytest.fixture(name="plot_patch")
def make_plot_patch(monkeypatch):
    plot_patch = PlotPatch()
    monkeypatch.setattr(pyplot, "savefig", plot_patch)
    return plot_patch



from __future__ import division, print_function

import autofit as af
import os

test_path = "{}/config/".format(os.path.dirname(os.path.realpath(__file__)))
af.conf.instance = af.conf.Config(config_path=test_path)

import numpy as np
import pytest


class TestGaussian:
    def test__transform_grid_wrapper_and_move_radial_minimum_wrappers(self):

        gaussian = aast.EllipticalGaussian(centre=(0.0, 0.0))

        image_1 = gaussian.profile_image_from_grid(grid=grid_01)
        image_0 = gaussian.profile_image_from_grid(grid=np.array([[0.000000001, 0.0]]))
        assert image_0 == pytest.approx(image_1, 1.0e-4)

        gaussian = aast.EllipticalGaussian(centre=(1.0, 1.0))

        image_1 = gaussian.profile_image_from_grid(grid=grid_22)
        image_0 = gaussian.profile_image_from_grid(grid=np.array([[1.000000001, 1.0]]))
        assert image_0 == pytest.approx(image_1, 1.0e-4)

        gaussian = aast.SphericalGaussian(centre=(0.0, 0.0))

        image_1 = gaussian.profile_image_from_grid(grid=grid_01)
        image_0 = gaussian.profile_image_from_grid(grid=np.array([[0.000000001, 0.0]]))
        assert image_0 == pytest.approx(image_1, 1.0e-4)

        gaussian = aast.SphericalGaussian(centre=(1.0, 1.0))

        image_1 = gaussian.profile_image_from_grid(grid=grid_22)
        image_0 = gaussian.profile_image_from_grid(grid=np.array([[1.000000001, 1.0]]))
        assert image_0 == pytest.approx(image_1, 1.0e-4)


class TestSersic:
    def test__transform_grid_wrapper_and_move_radial_minimum_wrappers(self):

        sersic = aast.EllipticalSersic(centre=(0.0, 0.0))

        image_1 = sersic.profile_image_from_grid(grid=grid_01)
        image_0 = sersic.profile_image_from_grid(grid=np.array([[0.000000001, 0.0]]))
        assert image_0 == pytest.approx(image_1, 1.0e-4)

        sersic = aast.EllipticalSersic(centre=(1.0, 1.0))

        image_1 = sersic.profile_image_from_grid(grid=grid_22)
        image_0 = sersic.profile_image_from_grid(grid=np.array([[1.000000001, 1.0]]))
        assert image_0 == pytest.approx(image_1, 1.0e-4)

        sersic = aast.light_profiles.SphericalSersic(centre=(0.0, 0.0))

        image_1 = sersic.profile_image_from_grid(grid=grid_01)
        image_0 = sersic.profile_image_from_grid(grid=np.array([[0.000000001, 0.0]]))
        assert image_0 == pytest.approx(image_1, 1.0e-4)

        sersic = aast.light_profiles.SphericalSersic(centre=(1.0, 1.0))

        image_1 = sersic.profile_image_from_grid(grid=grid_22)
        image_0 = sersic.profile_image_from_grid(grid=np.array([[1.000000001, 1.0]]))
        assert image_0 == pytest.approx(image_1, 1.0e-4)


class TestExponential:
    def test__transform_grid_wrapper_and_move_radial_minimum_wrappers(self):

        exponential = aast.EllipticalExponential(centre=(0.0, 0.0))

        image_1 = exponential.profile_image_from_grid(grid=grid_01)
        image_0 = exponential.profile_image_from_grid(
            grid=np.array([[0.000000001, 0.0]])
        )
        assert image_0 == pytest.approx(image_1, 1.0e-4)

        exponential = aast.EllipticalExponential(centre=(1.0, 1.0))

        image_1 = exponential.profile_image_from_grid(grid=grid_22)
        image_0 = exponential.profile_image_from_grid(
            grid=np.array([[1.000000001, 1.0]])
        )
        assert image_0 == pytest.approx(image_1, 1.0e-4)

        exponential = aast.SphericalExponential(centre=(0.0, 0.0))

        image_1 = exponential.profile_image_from_grid(grid=grid_01)
        image_0 = exponential.profile_image_from_grid(
            grid=np.array([[0.000000001, 0.0]])
        )
        assert image_0 == pytest.approx(image_1, 1.0e-4)

        exponential = aast.SphericalExponential(centre=(1.0, 1.0))

        image_1 = exponential.profile_image_from_grid(grid=grid_22)
        image_0 = exponential.profile_image_from_grid(
            grid=np.array([[1.000000001, 1.0]])
        )
        assert image_0 == pytest.approx(image_1, 1.0e-4)


class TestDevVaucouleurs:
    def test__transform_grid_wrapper_and_move_radial_minimum_wrappers(self):
        dev_vaucouleurs = aast.EllipticalDevVaucouleurs(centre=(0.0, 0.0))

        image_1 = dev_vaucouleurs.profile_image_from_grid(grid=grid_01)
        image_0 = dev_vaucouleurs.profile_image_from_grid(
            grid=np.array([[0.000000001, 0.0]])
        )
        assert image_0 == pytest.approx(image_1, 1.0e-4)

        dev_vaucouleurs = aast.EllipticalDevVaucouleurs(centre=(1.0, 1.0))

        image_1 = dev_vaucouleurs.profile_image_from_grid(grid=grid_22)
        image_0 = dev_vaucouleurs.profile_image_from_grid(
            grid=np.array([[1.000000001, 1.0]])
        )
        assert image_0 == pytest.approx(image_1, 1.0e-4)

        dev_vaucouleurs = aast.SphericalDevVaucouleurs(centre=(0.0, 0.0))

        image_1 = dev_vaucouleurs.profile_image_from_grid(grid=grid_01)
        image_0 = dev_vaucouleurs.profile_image_from_grid(
            grid=np.array([[0.000000001, 0.0]])
        )
        assert image_0 == pytest.approx(image_1, 1.0e-4)

        dev_vaucouleurs = aast.SphericalDevVaucouleurs(centre=(1.0, 1.0))

        image_1 = dev_vaucouleurs.profile_image_from_grid(grid=grid_22)
        image_0 = dev_vaucouleurs.profile_image_from_grid(
            grid=np.array([[1.000000001, 1.0]])
        )
        assert image_0 == pytest.approx(image_1, 1.0e-4)


class TestCoreSersic(object):
    def test__transform_grid_wrapper_and_move_radial_minimum_wrappers(self):
        core_sersic = aast.EllipticalCoreSersic(centre=(0.0, 0.0))

        image_1 = core_sersic.profile_image_from_grid(grid=grid_01)
        image_0 = core_sersic.profile_image_from_grid(
            grid=np.array([[0.000000001, 0.0]])
        )
        assert image_0 == pytest.approx(image_1, 1.0e-4)

        core_sersic = aast.EllipticalCoreSersic(centre=(1.0, 1.0))

        image_1 = core_sersic.profile_image_from_grid(grid=grid_22)
        image_0 = core_sersic.profile_image_from_grid(
            grid=np.array([[1.000000001, 1.0]])
        )
        assert image_0 == pytest.approx(image_1, 1.0e-4)

        core_sersic = aast.SphericalCoreSersic(centre=(0.0, 0.0))

        image_1 = core_sersic.profile_image_from_grid(grid=grid_01)
        image_0 = core_sersic.profile_image_from_grid(
            grid=np.array([[0.000000001, 0.0]])
        )
        assert image_0 == pytest.approx(image_1, 1.0e-4)

        core_sersic = aast.SphericalCoreSersic(centre=(1.0, 1.0))

        image_1 = core_sersic.profile_image_from_grid(grid=grid_22)
        image_0 = core_sersic.profile_image_from_grid(
            grid=np.array([[1.000000001, 1.0]])
        )
        assert image_0 == pytest.approx(image_1, 1.0e-4)


class TestPointMass(object):
    def test__transform_grid_wrapper_and_move_radial_minimum_wrappers(self):

        point_mass = aast.PointMass(centre=(0.0, 0.0), einstein_radius=1.0)

        deflections_1 = point_mass.deflections_from_grid(grid=grid_01)
        deflections_0 = point_mass.deflections_from_grid(
            grid=np.array([[0.00000001, 0.0]])
        )
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)

        point_mass = aast.PointMass(centre=(1.0, 1.0), einstein_radius=1.0)

        deflections_1 = point_mass.deflections_from_grid(grid=grid_22)
        deflections_0 = point_mass.deflections_from_grid(
            grid=np.array([[1.000000001, 1.0]])
        )
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)


class TestCoredPowerLaw(object):
    def test__transform_grid_wrapper_and_move_radial_minimum_wrappers(self):
        cored_power_law = aast.EllipticalCoredPowerLaw(
            centre=(0.0, 0.0), einstein_radius=1.0, slope=2.0
        )

        convergence_1 = cored_power_law.convergence_from_grid(grid=grid_01)
        convergence_0 = cored_power_law.convergence_from_grid(
            grid=np.array([[1e-8, 0.0]])
        )
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)


class TestPowerLaw(object):
    def test__transform_grid_wrapper_and_move_radial_minimum_wrappers(self):

        power_law = aast.EllipticalPowerLaw(
            centre=(0.0, 0.0), einstein_radius=1.0, slope=2.0
        )

        convergence_1 = power_law.convergence_from_grid(grid=grid_01)
        convergence_0 = power_law.convergence_from_grid(grid=np.array([[1.0e-9, 0.0]]))
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        power_law = aast.SphericalPowerLaw(
            centre=(-0.7, 0.5), einstein_radius=1.3, slope=1.8
        )

        mask = np.array(
            [
                [True, True, True, True, True],
                [True, False, False, False, True],
                [True, False, True, False, True],
                [True, False, False, False, True],
                [True, True, True, True, True],
            ]
        )

        mask = aa.mask.manual(mask, real_space_pixel_scales=1.0)

        grid = aa.masked.grid.from_mask(mask=mask)

        regular_with_interp = grid.new_grid_with_interpolator(
            pixel_scale_interpolation_grid=0.5
        )
        interp_deflections = power_law.deflections_from_grid(grid=regular_with_interp)

        interpolator = grids.Interpolator.from_mask_grid_and_pixel_scale_interpolation_grids(
            mask=mask, grid=grid, pixel_scale_interpolation_grid=0.5
        )

        interp_deflections_values = power_law.deflections_from_grid(
            grid=interpolator.interp_grid
        )

        interp_deflections_manual_y = interpolator.interpolated_values_from_values(
            values=interp_deflections_values[:, 0]
        )
        interp_deflections_manual_x = interpolator.interpolated_values_from_values(
            values=interp_deflections_values[:, 1]
        )

        assert (interp_deflections_manual_y != interp_deflections[:, 0]).all()
        assert (interp_deflections_manual_x != interp_deflections[:, 1]).all()


class TestCoredIsothermal(object):
    def test__transform_grid_wrapper_and_move_radial_minimum_wrappers(self):
        cored_isothermal = aast.EllipticalCoredIsothermal(
            centre=(0.0, 0.0), einstein_radius=1.0
        )

        convergence_1 = cored_isothermal.convergence_from_grid(grid=grid_01)
        convergence_0 = cored_isothermal.convergence_from_grid(
            grid=np.array([[0.000000001, 0.0]])
        )
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        cored_isothermal = aast.EllipticalCoredIsothermal(
            centre=(1.0, 1.0), einstein_radius=1.0
        )

        convergence_1 = cored_isothermal.convergence_from_grid(grid=grid_22)
        convergence_0 = cored_isothermal.convergence_from_grid(
            grid=np.array([[1.000000001, 1.0]])
        )
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        cored_isothermal = aast.SphericalCoredIsothermal(
            centre=(0.0, 0.0), einstein_radius=1.0
        )

        convergence_1 = cored_isothermal.convergence_from_grid(grid=grid_01)
        convergence_0 = cored_isothermal.convergence_from_grid(
            grid=np.array([[0.000000001, 0.0]])
        )
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        cored_isothermal = aast.SphericalCoredIsothermal(
            centre=(1.0, 1.0), einstein_radius=1.0
        )

        convergence_1 = cored_isothermal.convergence_from_grid(grid=grid_22)
        convergence_0 = cored_isothermal.convergence_from_grid(
            grid=np.array([[1.000000001, 1.0]])
        )
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        cored_isothermal = aast.EllipticalCoredIsothermal(
            centre=(0.0, 0.0), einstein_radius=1.0
        )

        potential_1 = cored_isothermal.potential_from_grid(grid=grid_01)
        potential_0 = cored_isothermal.potential_from_grid(
            grid=np.array([[0.000000001, 0.0]])
        )
        assert potential_0 == pytest.approx(potential_1, 1.0e-4)

        cored_isothermal = aast.EllipticalCoredIsothermal(
            centre=(1.0, 1.0), einstein_radius=1.0
        )

        potential_1 = cored_isothermal.potential_from_grid(grid=grid_22)
        potential_0 = cored_isothermal.potential_from_grid(
            grid=np.array([[1.000000001, 1.0]])
        )
        assert potential_0 == pytest.approx(potential_1, 1.0e-4)

        cored_isothermal = aast.SphericalCoredIsothermal(
            centre=(0.0, 0.0), einstein_radius=1.0
        )

        potential_1 = cored_isothermal.potential_from_grid(grid=grid_01)
        potential_0 = cored_isothermal.potential_from_grid(
            grid=np.array([[0.000000001, 0.0]])
        )
        assert potential_0 == pytest.approx(potential_1, 1.0e-4)

        cored_isothermal = aast.SphericalCoredIsothermal(
            centre=(1.0, 1.0), einstein_radius=1.0
        )

        potential_1 = cored_isothermal.potential_from_grid(grid=grid_22)
        potential_0 = cored_isothermal.potential_from_grid(
            grid=np.array([[1.000000001, 1.0]])
        )
        assert potential_0 == pytest.approx(potential_1, 1.0e-4)

        cored_isothermal = aast.EllipticalCoredIsothermal(
            centre=(0.0, 0.0), einstein_radius=1.0
        )

        deflections_1 = cored_isothermal.deflections_from_grid(grid=grid_01)
        deflections_0 = cored_isothermal.deflections_from_grid(
            grid=np.array([[0.000000001, 0.0]])
        )
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)

        cored_isothermal = aast.EllipticalCoredIsothermal(
            centre=(1.0, 1.0), einstein_radius=1.0
        )

        deflections_1 = cored_isothermal.deflections_from_grid(grid=grid_22)
        deflections_0 = cored_isothermal.deflections_from_grid(
            grid=np.array([[1.000000001, 1.0]])
        )
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)

        cored_isothermal = aast.SphericalCoredIsothermal(
            centre=(0.0, 0.0), einstein_radius=1.0
        )

        deflections_1 = cored_isothermal.deflections_from_grid(grid=grid_01)
        deflections_0 = cored_isothermal.deflections_from_grid(
            grid=np.array([[0.000000001, 0.0]])
        )
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)

        cored_isothermal = aast.SphericalCoredIsothermal(
            centre=(1.0, 1.0), einstein_radius=1.0
        )

        deflections_1 = cored_isothermal.deflections_from_grid(grid=grid_22)
        deflections_0 = cored_isothermal.deflections_from_grid(
            grid=np.array([[1.000000001, 1.0]])
        )
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)


class TestIsothermal(object):
    def test__transform_grid_wrapper_and_move_radial_minimum_wrappers(self):
        isothermal = aast.mass_profiles.EllipticalIsothermal(
            centre=(0.0, 0.0), einstein_radius=1.0
        )

        convergence_1 = isothermal.convergence_from_grid(grid=grid_01)
        convergence_0 = isothermal.convergence_from_grid(
            grid=np.array([[0.000000001, 0.0]])
        )
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        isothermal = aast.mass_profiles.EllipticalIsothermal(
            centre=(1.0, 1.0), einstein_radius=1.0
        )

        convergence_1 = isothermal.convergence_from_grid(grid=grid_22)
        convergence_0 = isothermal.convergence_from_grid(
            grid=np.array([[1.000000001, 1.0]])
        )
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        isothermal = aast.mass_profiles.SphericalIsothermal(
            centre=(0.0, 0.0), einstein_radius=1.0
        )

        convergence_1 = isothermal.convergence_from_grid(grid=grid_01)
        convergence_0 = isothermal.convergence_from_grid(
            grid=np.array([[0.000000001, 0.0]])
        )
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        isothermal = aast.mass_profiles.SphericalIsothermal(
            centre=(1.0, 1.0), einstein_radius=1.0
        )

        convergence_1 = isothermal.convergence_from_grid(grid=grid_22)
        convergence_0 = isothermal.convergence_from_grid(
            grid=np.array([[1.000000001, 1.0]])
        )
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        isothermal = aast.mass_profiles.EllipticalIsothermal(
            centre=(0.0, 0.0), einstein_radius=1.0
        )

        potential_1 = isothermal.potential_from_grid(grid=grid_01)
        potential_0 = isothermal.potential_from_grid(
            grid=np.array([[0.000000001, 0.0]])
        )
        assert potential_0 == pytest.approx(potential_1, 1.0e-4)

        isothermal = aast.mass_profiles.EllipticalIsothermal(
            centre=(1.0, 1.0), einstein_radius=1.0
        )

        potential_1 = isothermal.potential_from_grid(grid=grid_22)
        potential_0 = isothermal.potential_from_grid(
            grid=np.array([[1.000000001, 1.0]])
        )
        assert potential_0 == pytest.approx(potential_1, 1.0e-4)

        isothermal = aast.mass_profiles.SphericalIsothermal(
            centre=(0.0, 0.0), einstein_radius=1.0
        )

        potential_1 = isothermal.potential_from_grid(grid=grid_01)
        potential_0 = isothermal.potential_from_grid(
            grid=np.array([[0.000000001, 0.0]])
        )
        assert potential_0 == pytest.approx(potential_1, 1.0e-4)

        isothermal = aast.mass_profiles.SphericalIsothermal(
            centre=(1.0, 1.0), einstein_radius=1.0
        )

        potential_1 = isothermal.potential_from_grid(grid=grid_22)
        potential_0 = isothermal.potential_from_grid(
            grid=np.array([[1.000000001, 1.0]])
        )
        assert potential_0 == pytest.approx(potential_1, 1.0e-4)

        isothermal = aast.mass_profiles.EllipticalIsothermal(
            centre=(0.0, 0.0), einstein_radius=1.0
        )

        deflections_1 = isothermal.deflections_from_grid(grid=grid_01)
        deflections_0 = isothermal.deflections_from_grid(
            grid=np.array([[0.000000001, 0.0]])
        )
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)

        isothermal = aast.mass_profiles.EllipticalIsothermal(
            centre=(1.0, 1.0), einstein_radius=1.0
        )

        deflections_1 = isothermal.deflections_from_grid(grid=grid_22)
        deflections_0 = isothermal.deflections_from_grid(
            grid=np.array([[1.000000001, 1.0]])
        )
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)

        isothermal = aast.mass_profiles.SphericalIsothermal(
            centre=(0.0, 0.0), einstein_radius=1.0
        )

        deflections_1 = isothermal.deflections_from_grid(grid=grid_01)
        deflections_0 = isothermal.deflections_from_grid(
            grid=np.array([[0.000000001, 0.0]])
        )
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)

        isothermal = aast.mass_profiles.SphericalIsothermal(
            centre=(1.0, 1.0), einstein_radius=1.0
        )

        deflections_1 = isothermal.deflections_from_grid(grid=grid_22)
        deflections_0 = isothermal.deflections_from_grid(
            grid=np.array([[1.000000001, 1.0]])
        )
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)


class TestGeneralizedNFW(object):
    def test__transform_grid_wrapper_and_move_radial_minimum_wrappers(self):
        gnfw = aast.mass_profiles.SphericalGeneralizedNFW(centre=(0.0, 0.0))

        convergence_1 = gnfw.convergence_from_grid(grid=grid_01)
        convergence_0 = gnfw.convergence_from_grid(grid=np.array([[0.000000001, 0.0]]))
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        gnfw = aast.mass_profiles.SphericalGeneralizedNFW(centre=(1.0, 1.0))

        convergence_1 = gnfw.convergence_from_grid(grid=grid_22)
        convergence_0 = gnfw.convergence_from_grid(grid=np.array([[1.000000001, 1.0]]))
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        gnfw = aast.mass_profiles.SphericalGeneralizedNFW(centre=(0.0, 0.0))

        potential_1 = gnfw.potential_from_grid(grid=grid_01)
        potential_0 = gnfw.potential_from_grid(grid=np.array([[0.000000001, 0.0]]))
        assert potential_0 == pytest.approx(potential_1, 1.0e-4)

        gnfw = aast.mass_profiles.SphericalGeneralizedNFW(centre=(1.0, 1.0))

        potential_1 = gnfw.potential_from_grid(grid=grid_22)
        potential_0 = gnfw.potential_from_grid(grid=np.array([[1.000000001, 1.0]]))
        assert potential_0 == pytest.approx(potential_1, 1.0e-4)

        gnfw = aast.mass_profiles.SphericalGeneralizedNFW(centre=(0.0, 0.0))

        deflections_1 = gnfw.deflections_from_grid(grid=grid_01)
        deflections_0 = gnfw.deflections_from_grid(grid=np.array([[0.000000001, 0.0]]))
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)

        gnfw = aast.mass_profiles.SphericalGeneralizedNFW(centre=(1.0, 1.0))

        deflections_1 = gnfw.deflections_from_grid(grid=grid_22)
        deflections_0 = gnfw.deflections_from_grid(grid=np.array([[1.000000001, 1.0]]))
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)


class TestTruncatedNFW(object):
    def test__transform_grid_wrapper_and_move_radial_minimum_wrappers(self):
        truncated_nfw = aast.mass_profiles.SphericalTruncatedNFW(centre=(0.0, 0.0))

        convergence_1 = truncated_nfw.convergence_from_grid(grid=grid_01)
        convergence_0 = truncated_nfw.convergence_from_grid(
            grid=np.array([[0.000000001, 0.0]])
        )
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        truncated_nfw = aast.mass_profiles.SphericalTruncatedNFW(centre=(1.0, 1.0))

        convergence_1 = truncated_nfw.convergence_from_grid(grid=grid_22)
        convergence_0 = truncated_nfw.convergence_from_grid(
            grid=np.array([[1.000000001, 1.0]])
        )
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        truncated_nfw = aast.mass_profiles.SphericalTruncatedNFW(centre=(0.0, 0.0))

        potential_1 = truncated_nfw.potential_from_grid(grid=grid_01)
        potential_0 = truncated_nfw.potential_from_grid(
            grid=np.array([[0.000000001, 0.0]])
        )
        assert potential_0 == pytest.approx(potential_1, 1.0e-4)

        truncated_nfw = aast.mass_profiles.SphericalTruncatedNFW(centre=(1.0, 1.0))

        potential_1 = truncated_nfw.potential_from_grid(grid=grid_22)
        potential_0 = truncated_nfw.potential_from_grid(
            grid=np.array([[1.000000001, 1.0]])
        )
        assert potential_0 == pytest.approx(potential_1, 1.0e-4)

        truncated_nfw = aast.mass_profiles.SphericalTruncatedNFW(centre=(0.0, 0.0))

        deflections_1 = truncated_nfw.deflections_from_grid(grid=grid_01)
        deflections_0 = truncated_nfw.deflections_from_grid(
            grid=np.array([[0.000000001, 0.0]])
        )
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)

        truncated_nfw = aast.mass_profiles.SphericalTruncatedNFW(centre=(1.0, 1.0))

        deflections_1 = truncated_nfw.deflections_from_grid(grid=grid_22)
        deflections_0 = truncated_nfw.deflections_from_grid(
            grid=np.array([[1.000000001, 1.0]])
        )
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)


class TestNFW(object):
    def test__transform_grid_wrapper_and_move_radial_minimum_wrappers(self):
        nfw = aast.mass_profiles.EllipticalNFW(centre=(0.0, 0.0))

        convergence_1 = nfw.convergence_from_grid(grid=grid_01)
        convergence_0 = nfw.convergence_from_grid(grid=np.array([[0.000000001, 0.0]]))
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        nfw = aast.mass_profiles.EllipticalNFW(centre=(1.0, 1.0))

        convergence_1 = nfw.convergence_from_grid(grid=grid_22)
        convergence_0 = nfw.convergence_from_grid(grid=np.array([[1.000000001, 1.0]]))
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        nfw = aast.mass_profiles.SphericalNFW(centre=(0.0, 0.0))

        convergence_1 = nfw.convergence_from_grid(grid=grid_01)
        convergence_0 = nfw.convergence_from_grid(grid=np.array([[0.000000001, 0.0]]))
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        nfw = aast.mass_profiles.SphericalNFW(centre=(1.0, 1.0))

        convergence_1 = nfw.convergence_from_grid(grid=grid_22)
        convergence_0 = nfw.convergence_from_grid(grid=np.array([[1.000000001, 1.0]]))
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        nfw = aast.mass_profiles.EllipticalNFW(centre=(0.0, 0.0))

        potential_1 = nfw.potential_from_grid(grid=grid_01)
        potential_0 = nfw.potential_from_grid(grid=np.array([[0.000000001, 0.0]]))
        assert potential_0 == pytest.approx(potential_1, 1.0e-4)

        nfw = aast.mass_profiles.EllipticalNFW(centre=(1.0, 1.0))

        potential_1 = nfw.potential_from_grid(grid=grid_22)
        potential_0 = nfw.potential_from_grid(grid=np.array([[1.000000001, 1.0]]))
        assert potential_0 == pytest.approx(potential_1, 1.0e-4)

        nfw = aast.mass_profiles.SphericalNFW(centre=(0.0, 0.0))

        potential_1 = nfw.potential_from_grid(grid=grid_01)
        potential_0 = nfw.potential_from_grid(grid=np.array([[0.000000001, 0.0]]))
        assert potential_0 == pytest.approx(potential_1, 1.0e-4)

        nfw = aast.mass_profiles.SphericalNFW(centre=(1.0, 1.0))

        potential_1 = nfw.potential_from_grid(grid=grid_22)
        potential_0 = nfw.potential_from_grid(grid=np.array([[1.000000001, 1.0]]))
        assert potential_0 == pytest.approx(potential_1, 1.0e-4)

        nfw = aast.mass_profiles.EllipticalNFW(centre=(0.0, 0.0))

        deflections_1 = nfw.deflections_from_grid(grid=grid_01)
        deflections_0 = nfw.deflections_from_grid(grid=np.array([[0.000000001, 0.0]]))
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)

        nfw = aast.mass_profiles.EllipticalNFW(centre=(1.0, 1.0))

        deflections_1 = nfw.deflections_from_grid(grid=grid_22)
        deflections_0 = nfw.deflections_from_grid(grid=np.array([[1.000000001, 1.0]]))
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)

        nfw = aast.mass_profiles.SphericalNFW(centre=(0.0, 0.0))

        deflections_1 = nfw.deflections_from_grid(grid=grid_01)
        deflections_0 = nfw.deflections_from_grid(grid=np.array([[0.000000001, 0.0]]))
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)

        nfw = aast.mass_profiles.SphericalNFW(centre=(1.0, 1.0))

        deflections_1 = nfw.deflections_from_grid(grid=grid_22)
        deflections_0 = nfw.deflections_from_grid(grid=np.array([[1.000000001, 1.0]]))
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)


class TestSersic(object):
    def test__transform_grid_wrapper_and_move_radial_minimum_wrappers(self):
        sersic = aast.EllipticalSersic(centre=(0.0, 0.0))

        convergence_1 = sersic.convergence_from_grid(grid=grid_01)
        convergence_0 = sersic.convergence_from_grid(
            grid=np.array([[0.000000001, 0.0]])
        )
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        sersic = aast.EllipticalSersic(centre=(1.0, 1.0))

        convergence_1 = sersic.convergence_from_grid(grid=grid_22)
        convergence_0 = sersic.convergence_from_grid(
            grid=np.array([[1.000000001, 1.0]])
        )
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        sersic = aast.SphericalSersic(centre=(0.0, 0.0))

        convergence_1 = sersic.convergence_from_grid(grid=grid_01)
        convergence_0 = sersic.convergence_from_grid(
            grid=np.array([[0.000000001, 0.0]])
        )
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        sersic = aast.SphericalSersic(centre=(1.0, 1.0))

        convergence_1 = sersic.convergence_from_grid(grid=grid_22)
        convergence_0 = sersic.convergence_from_grid(
            grid=np.array([[1.000000001, 1.0]])
        )
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        sersic = aast.EllipticalSersic(centre=(0.0, 0.0))

        deflections_1 = sersic.deflections_from_grid(grid=grid_01)
        deflections_0 = sersic.deflections_from_grid(
            grid=np.array([[0.000000001, 0.0]])
        )
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)

        sersic = aast.EllipticalSersic(centre=(1.0, 1.0))

        deflections_1 = sersic.deflections_from_grid(grid=grid_22)
        deflections_0 = sersic.deflections_from_grid(
            grid=np.array([[1.000000001, 1.0]])
        )
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)

        sersic = aast.SphericalSersic(centre=(0.0, 0.0))

        deflections_1 = sersic.deflections_from_grid(grid=grid_01)
        deflections_0 = sersic.deflections_from_grid(
            grid=np.array([[0.000000001, 0.0]])
        )
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)

        sersic = aast.SphericalSersic(centre=(1.0, 1.0))

        deflections_1 = sersic.deflections_from_grid(grid=grid_22)
        deflections_0 = sersic.deflections_from_grid(
            grid=np.array([[1.000000001, 1.0]])
        )
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)


class TestExponential(object):
    def test__transform_grid_wrapper_and_move_radial_minimum_wrappers(self):
        exponential = aast.EllipticalExponential(centre=(0.0, 0.0))

        convergence_1 = exponential.convergence_from_grid(grid=grid_01)
        convergence_0 = exponential.convergence_from_grid(
            grid=np.array([[0.000000001, 0.0]])
        )
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        exponential = aast.EllipticalExponential(centre=(1.0, 1.0))

        convergence_1 = exponential.convergence_from_grid(grid=grid_22)
        convergence_0 = exponential.convergence_from_grid(
            grid=np.array([[1.000000001, 1.0]])
        )
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        exponential = aast.SphericalExponential(centre=(0.0, 0.0))

        convergence_1 = exponential.convergence_from_grid(grid=grid_01)
        convergence_0 = exponential.convergence_from_grid(
            grid=np.array([[0.000000001, 0.0]])
        )
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        exponential = aast.SphericalExponential(centre=(1.0, 1.0))

        convergence_1 = exponential.convergence_from_grid(grid=grid_22)
        convergence_0 = exponential.convergence_from_grid(
            grid=np.array([[1.000000001, 1.0]])
        )
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        exponential = aast.EllipticalExponential(centre=(0.0, 0.0))

        deflections_1 = exponential.deflections_from_grid(grid=grid_01)
        deflections_0 = exponential.deflections_from_grid(
            grid=np.array([[0.000000001, 0.0]])
        )
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)

        exponential = aast.EllipticalExponential(centre=(1.0, 1.0))

        deflections_1 = exponential.deflections_from_grid(grid=grid_22)
        deflections_0 = exponential.deflections_from_grid(
            grid=np.array([[1.000000001, 1.0]])
        )
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)

        exponential = aast.SphericalExponential(centre=(0.0, 0.0))

        deflections_1 = exponential.deflections_from_grid(grid=grid_01)
        deflections_0 = exponential.deflections_from_grid(
            grid=np.array([[0.000000001, 0.0]])
        )
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)

        exponential = aast.SphericalExponential(centre=(1.0, 1.0))

        deflections_1 = exponential.deflections_from_grid(grid=grid_22)
        deflections_0 = exponential.deflections_from_grid(
            grid=np.array([[1.000000001, 1.0]])
        )
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)


class TestDevVaucouleurs(object):
    def test__transform_grid_wrapper_and_move_radial_minimum_wrappers(self):
        dev_vaucouleurs = aast.EllipticalDevVaucouleurs(centre=(0.0, 0.0))

        convergence_1 = dev_vaucouleurs.convergence_from_grid(grid=grid_01)
        convergence_0 = dev_vaucouleurs.convergence_from_grid(
            grid=np.array([[0.000000001, 0.0]])
        )
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        dev_vaucouleurs = aast.EllipticalDevVaucouleurs(centre=(1.0, 1.0))

        convergence_1 = dev_vaucouleurs.convergence_from_grid(grid=grid_22)
        convergence_0 = dev_vaucouleurs.convergence_from_grid(
            grid=np.array([[1.000000001, 1.0]])
        )
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        dev_vaucouleurs = aast.SphericalDevVaucouleurs(centre=(0.0, 0.0))

        convergence_1 = dev_vaucouleurs.convergence_from_grid(grid=grid_01)
        convergence_0 = dev_vaucouleurs.convergence_from_grid(
            grid=np.array([[0.000000001, 0.0]])
        )
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        dev_vaucouleurs = aast.SphericalDevVaucouleurs(centre=(1.0, 1.0))

        convergence_1 = dev_vaucouleurs.convergence_from_grid(grid=grid_22)
        convergence_0 = dev_vaucouleurs.convergence_from_grid(
            grid=np.array([[1.000000001, 1.0]])
        )
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        dev_vaucouleurs = aast.EllipticalDevVaucouleurs(centre=(0.0, 0.0))

        deflections_1 = dev_vaucouleurs.deflections_from_grid(grid=grid_01)
        deflections_0 = dev_vaucouleurs.deflections_from_grid(
            grid=np.array([[0.000000001, 0.0]])
        )
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)

        dev_vaucouleurs = aast.EllipticalDevVaucouleurs(centre=(1.0, 1.0))

        deflections_1 = dev_vaucouleurs.deflections_from_grid(grid=grid_22)
        deflections_0 = dev_vaucouleurs.deflections_from_grid(
            grid=np.array([[1.000000001, 1.0]])
        )
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)

        dev_vaucouleurs = aast.SphericalDevVaucouleurs(centre=(0.0, 0.0))

        deflections_1 = dev_vaucouleurs.deflections_from_grid(grid=grid_01)
        deflections_0 = dev_vaucouleurs.deflections_from_grid(
            grid=np.array([[0.000000001, 0.0]])
        )
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)

        dev_vaucouleurs = aast.SphericalDevVaucouleurs(centre=(1.0, 1.0))

        deflections_1 = dev_vaucouleurs.deflections_from_grid(grid=grid_22)
        deflections_0 = dev_vaucouleurs.deflections_from_grid(
            grid=np.array([[1.000000001, 1.0]])
        )
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)


class TestSersicMassRadialGradient(object):
    def test__transform_grid_wrapper_and_move_radial_minimum_wrappers(self):
        sersic = aast.EllipticalSersicRadialGradient(centre=(0.0, 0.0))

        convergence_1 = sersic.convergence_from_grid(grid=grid_01)
        convergence_0 = sersic.convergence_from_grid(
            grid=np.array([[0.000000001, 0.0]])
        )
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        sersic = aast.EllipticalSersicRadialGradient(centre=(1.0, 1.0))

        convergence_1 = sersic.convergence_from_grid(grid=grid_22)
        convergence_0 = sersic.convergence_from_grid(
            grid=np.array([[1.000000001, 1.0]])
        )
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        sersic = aast.SphericalSersicRadialGradient(centre=(0.0, 0.0))

        convergence_1 = sersic.convergence_from_grid(grid=grid_01)
        convergence_0 = sersic.convergence_from_grid(
            grid=np.array([[0.000000001, 0.0]])
        )
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        sersic = aast.SphericalSersicRadialGradient(centre=(1.0, 1.0))

        convergence_1 = sersic.convergence_from_grid(grid=grid_22)
        convergence_0 = sersic.convergence_from_grid(
            grid=np.array([[1.000000001, 1.0]])
        )
        assert convergence_0 == pytest.approx(convergence_1, 1.0e-4)

        sersic = aast.EllipticalSersicRadialGradient(centre=(0.0, 0.0))

        deflections_1 = sersic.deflections_from_grid(grid=grid_01)
        deflections_0 = sersic.deflections_from_grid(
            grid=np.array([[0.000000001, 0.0]])
        )
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)

        sersic = aast.EllipticalSersicRadialGradient(centre=(1.0, 1.0))

        deflections_1 = sersic.deflections_from_grid(grid=grid_22)
        deflections_0 = sersic.deflections_from_grid(
            grid=np.array([[1.000000001, 1.0]])
        )
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)

        sersic = aast.SphericalSersicRadialGradient(centre=(0.0, 0.0))

        deflections_1 = sersic.deflections_from_grid(grid=grid_01)
        deflections_0 = sersic.deflections_from_grid(
            grid=np.array([[0.000000001, 0.0]])
        )
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)

        sersic = aast.SphericalSersicRadialGradient(centre=(1.0, 1.0))

        deflections_1 = sersic.deflections_from_grid(grid=grid_22)
        deflections_0 = sersic.deflections_from_grid(
            grid=np.array([[1.000000001, 1.0]])
        )
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)
        sersic = aast.SphericalSersicRadialGradient(
            centre=(-0.7, 0.5),
            intensity=5.0,
            effective_radius=0.2,
            sersic_index=2.0,
            mass_to_light_ratio=1.0,
            mass_to_light_gradient=1.5,
        )

        mask = np.array(
            [
                [True, True, True, True, True],
                [True, False, False, False, True],
                [True, False, True, False, True],
                [True, False, False, False, True],
                [True, True, True, True, True],
            ]
        )

        mask = aa.mask.manual(mask, real_space_pixel_scales=1.0)

        grid = aa.masked.grid.from_mask(mask=mask)

        regular_with_interp = grid.new_grid_with_interpolator(
            pixel_scale_interpolation_grid=0.5
        )
        interp_deflections = sersic.deflections_from_grid(grid=regular_with_interp)

        interpolator = grids.Interpolator.from_mask_grid_and_pixel_scale_interpolation_grids(
            mask=mask, grid=grid, pixel_scale_interpolation_grid=0.5
        )

        interp_deflections_values = sersic.deflections_from_grid(
            grid=interpolator.interp_grid
        )

        interp_deflections_manual_y = interpolator.interpolated_values_from_values(
            values=interp_deflections_values[:, 0]
        )
        interp_deflections_manual_x = interpolator.interpolated_values_from_values(
            values=interp_deflections_values[:, 1]
        )

        assert (interp_deflections_manual_y == interp_deflections[:, 0]).all()
        assert (interp_deflections_manual_x == interp_deflections[:, 1]).all()


class TestMassSheet(object):
    def test__transform_grid_wrapper_and_move_radial_minimum_wrappers(self):
        mass_sheet = aast.mass_profiles.MassSheet(centre=(0.0, 0.0))

        deflections_1 = mass_sheet.deflections_from_grid(grid=grid_01)
        deflections_0 = mass_sheet.deflections_from_grid(
            grid=np.array([[0.000000001, 0.0]])
        )
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)

        mass_sheet = aast.mass_profiles.MassSheet(centre=(1.0, 1.0))

        deflections_1 = mass_sheet.deflections_from_grid(grid=grid_22)
        deflections_0 = mass_sheet.deflections_from_grid(
            grid=np.array([[1.000000001, 1.0]])
        )
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)


class TestExternalShear(object):
    def test__transform_grid_wrapper_and_move_radial_minimum_wrappers(self):
        shear = aast.mass_profiles.ExternalShear(magnitude=0.1, phi=45.0)

        deflections_1 = shear.deflections_from_grid(grid=np.array([[1e-8, 0.0]]))
        deflections_0 = shear.deflections_from_grid(grid=np.array([[1e-9, 0.0]]))
        assert deflections_0 == pytest.approx(deflections_1, 1.0e-4)



import autoarray as aa
import matplotlib

backend = aa.conf.get_matplotlib_backend()
matplotlib.use(backend)
from matplotlib import pyplot as plt

from autoarray.util import plotter_util
from autoastro.plotters import profile_plotters, lens_plotter_util


def profile_image(
    galaxy,
    grid,
    mask=None,
    positions=None,
    include_critical_curves=False,
    include_caustics=False,
    as_subplot=False,
    plot_in_kpc=False,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Galaxy Profile Image",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    position_pointsize=10,
    grid_pointsize=1,
    output_path=None,
    output_format="show",
    output_filename="galaxy_profile_image",
):
    """Plot the image (e.g. the datas) of a galaxy, on a grid of (y,x) coordinates.

    Set *autoastro.datas.arrays.plotters.array_plotters* for a description of all innput parameters not described below.

    Parameters
    -----------
    galaxy : model.galaxy.aast.Galaxy
        The galaxy whose image are plotted.
    grid : ndarray or datas.arrays.grid_stacks.Grid
        The (y,x) coordinates of the grid, in an arrays of shape (total_coordinates, 2)
    """
    image = galaxy.profile_image_from_grid(grid=grid)

    lines = lens_plotter_util.get_critical_curves_and_caustics_from_lensing_object(
        obj=galaxy,
        include_critical_curves=include_critical_curves,
        include_caustics=include_caustics,
    )

    unit_label, unit_conversion_factor = lens_plotter_util.get_unit_label_and_unit_conversion_factor(
        obj=galaxy, plot_in_kpc=plot_in_kpc
    )

    aa.plot.array(
        array=image,
        mask=mask,
        points=positions,
        lines=lines,
        as_subplot=as_subplot,
        unit_label=unit_label,
        unit_conversion_factor=unit_conversion_factor,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        point_pointsize=position_pointsize,
        grid_pointsize=grid_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def convergence(
    galaxy,
    grid,
    mask=None,
    positions=None,
    include_critical_curves=False,
    include_caustics=False,
    as_subplot=False,
    plot_in_kpc=False,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Galaxy Convergence",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    position_pointsize=10,
    grid_pointsize=1,
    output_path=None,
    output_format="show",
    output_filename="galaxy_convergence",
):
    """Plot the convergence of a galaxy, on a grid of (y,x) coordinates.

    Set *autoastro.datas.arrays.plotters.array_plotters* for a description of all innput parameters not described below.

    Parameters
    -----------
    galaxy : model.galaxy.aast.Galaxy
        The galaxy whose convergence is plotted.
    grid : ndarray or datas.arrays.grid_stacks.Grid
        The (y,x) coordinates of the grid, in an arrays of shape (total_coordinates, 2)
    """
    convergence = galaxy.convergence_from_grid(grid=grid)

    lines = lens_plotter_util.get_critical_curves_and_caustics_from_lensing_object(
        obj=galaxy,
        include_critical_curves=include_critical_curves,
        include_caustics=include_caustics,
    )

    unit_label, unit_conversion_factor = lens_plotter_util.get_unit_label_and_unit_conversion_factor(
        obj=galaxy, plot_in_kpc=plot_in_kpc
    )

    aa.plot.array(
        array=convergence,
        mask=mask,
        points=positions,
        lines=lines,
        as_subplot=as_subplot,
        unit_label=unit_label,
        unit_conversion_factor=unit_conversion_factor,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        point_pointsize=position_pointsize,
        grid_pointsize=grid_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def potential(
    galaxy,
    grid,
    mask=None,
    positions=None,
    include_critical_curves=False,
    include_caustics=False,
    as_subplot=False,
    plot_in_kpc=False,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Galaxy Potential",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    position_pointsize=10,
    grid_pointsize=1,
    output_path=None,
    output_format="show",
    output_filename="galaxy_potential",
):
    """Plot the potential of a galaxy, on a grid of (y,x) coordinates.

     Set *autoastro.datas.arrays.plotters.array_plotters* for a description of all innput parameters not described below.

     Parameters
     -----------
    galaxy : model.galaxy.aast.Galaxy
         The galaxy whose potential is plotted.
    grid : ndarray or datas.arrays.grid_stacks.Grid
         The (y,x) coordinates of the grid, in an arrays of shape (total_coordinates, 2)
     """
    potential = galaxy.potential_from_grid(grid=grid)

    lines = lens_plotter_util.get_critical_curves_and_caustics_from_lensing_object(
        obj=galaxy,
        include_critical_curves=include_critical_curves,
        include_caustics=include_caustics,
    )

    unit_label, unit_conversion_factor = lens_plotter_util.get_unit_label_and_unit_conversion_factor(
        obj=galaxy, plot_in_kpc=plot_in_kpc
    )

    aa.plot.array(
        array=potential,
        mask=mask,
        points=positions,
        lines=lines,
        as_subplot=as_subplot,
        unit_label=unit_label,
        unit_conversion_factor=unit_conversion_factor,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        point_pointsize=position_pointsize,
        grid_pointsize=grid_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def deflections_y(
    galaxy,
    grid,
    mask=None,
    positions=None,
    include_critical_curves=False,
    include_caustics=False,
    as_subplot=False,
    plot_in_kpc=False,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Galaxy Deflections (y)",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    position_pointsize=10,
    grid_pointsize=1,
    output_path=None,
    output_format="show",
    output_filename="galaxy_deflections_y",
):
    """Plot the y component of the deflection angles of a galaxy, on a grid of (y,x) coordinates.

    Set *autoastro.datas.arrays.plotters.array_plotters* for a description of all innput parameters not described below.

    Parameters
    -----------
    galaxy : model.galaxy.aast.Galaxy
        The galaxy whose y deflecton angles are plotted.
    grid : ndarray or datas.arrays.grid_stacks.Grid
        The (y,x) coordinates of the grid, in an arrays of shape (total_coordinates, 2)
    """
    deflections = galaxy.deflections_from_grid(grid=grid)
    deflections_y = grid.mapping.array_from_sub_array_1d(sub_array_1d=deflections[:, 0])

    lines = lens_plotter_util.get_critical_curves_and_caustics_from_lensing_object(
        obj=galaxy,
        include_critical_curves=include_critical_curves,
        include_caustics=include_caustics,
    )

    unit_label, unit_conversion_factor = lens_plotter_util.get_unit_label_and_unit_conversion_factor(
        obj=galaxy, plot_in_kpc=plot_in_kpc
    )

    aa.plot.array(
        array=deflections_y,
        mask=mask,
        points=positions,
        lines=lines,
        as_subplot=as_subplot,
        unit_label=unit_label,
        unit_conversion_factor=unit_conversion_factor,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        point_pointsize=position_pointsize,
        grid_pointsize=grid_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def deflections_x(
    galaxy,
    grid,
    mask=None,
    positions=None,
    include_critical_curves=False,
    include_caustics=False,
    as_subplot=False,
    plot_in_kpc=False,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Galaxy Deflections (x)",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    position_pointsize=10,
    grid_pointsize=1,
    output_path=None,
    output_format="show",
    output_filename="galaxy_deflections_x",
):
    """Plot the x component of the deflection angles of a galaxy, on a grid of (y,x) coordinates.

     Set *autoastro.datas.arrays.plotters.array_plotters* for a description of all innput parameters not described below.

     Parameters
     -----------
    galaxy : model.galaxy.aast.Galaxy
         The galaxy whose x deflecton angles are plotted.
     grid : ndarray or datas.arrays.grid_stacks.Grid
         The (y,x) coordinates of the grid, in an arrays of shape (total_coordinates, 2)
     """
    deflections = galaxy.deflections_from_grid(grid=grid)
    deflections_x = grid.mapping.array_from_sub_array_1d(sub_array_1d=deflections[:, 1])

    lines = lens_plotter_util.get_critical_curves_and_caustics_from_lensing_object(
        obj=galaxy,
        include_critical_curves=include_critical_curves,
        include_caustics=include_caustics,
    )

    unit_label, unit_conversion_factor = lens_plotter_util.get_unit_label_and_unit_conversion_factor(
        obj=galaxy, plot_in_kpc=plot_in_kpc
    )

    aa.plot.array(
        array=deflections_x,
        mask=mask,
        points=positions,
        lines=lines,
        as_subplot=as_subplot,
        unit_label=unit_label,
        unit_conversion_factor=unit_conversion_factor,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        point_pointsize=position_pointsize,
        grid_pointsize=grid_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def magnification(
    galaxy,
    grid,
    mask=None,
    positions=None,
    include_critical_curves=False,
    include_caustics=False,
    as_subplot=False,
    plot_in_kpc=False,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Galaxy Magnification",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    position_pointsize=10,
    grid_pointsize=1,
    output_path=None,
    output_format="show",
    output_filename="galaxy_magnification",
):
    """Plot the magnification of a galaxy, on a grid of (y,x) coordinates.

     Set *autoastro.datas.arrays.plotters.array_plotters* for a description of all innput parameters not described below.

     Parameters
     -----------
    galaxy : model.galaxy.aast.Galaxy
         The galaxy whose magnification is plotted.
    grid : ndarray or datas.arrays.grid_stacks.Grid
         The (y,x) coordinates of the grid, in an arrays of shape (total_coordinates, 2)
     """
    magnification = galaxy.magnification_from_grid(grid=grid)

    lines = lens_plotter_util.get_critical_curves_and_caustics_from_lensing_object(
        obj=galaxy,
        include_critical_curves=include_critical_curves,
        include_caustics=include_caustics,
    )

    unit_label, unit_conversion_factor = lens_plotter_util.get_unit_label_and_unit_conversion_factor(
        obj=galaxy, plot_in_kpc=plot_in_kpc
    )

    aa.plot.array(
        array=magnification,
        mask=mask,
        points=positions,
        lines=lines,
        as_subplot=as_subplot,
        unit_label=unit_label,
        unit_conversion_factor=unit_conversion_factor,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        point_pointsize=position_pointsize,
        grid_pointsize=grid_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def profile_image_subplot(
    galaxy,
    grid,
    mask=None,
    positions=None,
    plot_in_kpc=False,
    figsize=None,
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    titlesize=10,
    xlabelsize=10,
    ylabelsize=10,
    xyticksize=10,
    mask_pointsize=10,
    position_pointsize=10,
    grid_pointsize=1,
    output_path=None,
    output_format="show",
    output_filename="galaxy_individual_image",
):

    total_light_profiles = len(galaxy.light_profiles)
    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(
        number_subplots=total_light_profiles
    )

    if figsize is None:
        figsize = figsize_tool

    plt.figure(figsize=figsize)

    unit_label, unit_conversion_factor = lens_plotter_util.get_unit_label_and_unit_conversion_factor(
        obj=galaxy, plot_in_kpc=plot_in_kpc
    )

    for i, light_profile in enumerate(galaxy.light_profiles):

        plt.subplot(rows, columns, i + 1)

        profile_plotters.image(
            light_profile=light_profile,
            mask=mask,
            positions=positions,
            grid=grid,
            as_subplot=True,
            unit_label=unit_label,
            unit_conversion_factor=unit_conversion_factor,
            figsize=figsize,
            aspect=aspect,
            cmap=cmap,
            norm=norm,
            norm_min=norm_min,
            norm_max=norm_max,
            linthresh=linthresh,
            linscale=linscale,
            cb_ticksize=cb_ticksize,
            cb_fraction=cb_fraction,
            cb_pad=cb_pad,
            cb_tick_values=cb_tick_values,
            cb_tick_labels=cb_tick_labels,
            title="Galaxy Component",
            titlesize=titlesize,
            xlabelsize=xlabelsize,
            ylabelsize=ylabelsize,
            xyticksize=xyticksize,
            mask_pointsize=mask_pointsize,
            position_pointsize=position_pointsize,
            grid_pointsize=grid_pointsize,
            output_path=output_path,
            output_format=output_format,
            output_filename=output_filename,
        )

    plotter_util.output_subplot_array(
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )
    plt.close()


def convergence_subplot(
    galaxy,
    grid,
    mask=None,
    positions=None,
    plot_in_kpc=False,
    figsize=None,
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    titlesize=10,
    xlabelsize=10,
    ylabelsize=10,
    xyticksize=10,
    mask_pointsize=10,
    position_pointsize=10,
    grid_pointsize=1,
    output_path=None,
    output_format="show",
    output_filename="galaxy_individual_convergence",
):

    total_mass_profiles = len(galaxy.mass_profiles)
    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(
        number_subplots=total_mass_profiles
    )

    if figsize is None:
        figsize = figsize_tool

    plt.figure(figsize=figsize)

    unit_label, unit_conversion_factor = lens_plotter_util.get_unit_label_and_unit_conversion_factor(
        obj=galaxy, plot_in_kpc=plot_in_kpc
    )

    for i, mass_profile in enumerate(galaxy.mass_profiles):

        plt.subplot(rows, columns, i + 1)

        profile_plotters.convergence(
            mass_profile=mass_profile,
            grid=grid,
            mask=mask,
            positions=positions,
            as_subplot=True,
            unit_label=unit_label,
            unit_conversion_factor=unit_conversion_factor,
            figsize=figsize,
            aspect=aspect,
            cmap=cmap,
            norm=norm,
            norm_min=norm_min,
            norm_max=norm_max,
            linthresh=linthresh,
            linscale=linscale,
            cb_ticksize=cb_ticksize,
            cb_fraction=cb_fraction,
            cb_pad=cb_pad,
            cb_tick_values=cb_tick_values,
            cb_tick_labels=cb_tick_labels,
            title="Galaxy Component",
            titlesize=titlesize,
            xlabelsize=xlabelsize,
            ylabelsize=ylabelsize,
            xyticksize=xyticksize,
            mask_pointsize=mask_pointsize,
            position_pointsize=position_pointsize,
            grid_pointsize=grid_pointsize,
            output_path=output_path,
            output_format=output_format,
            output_filename=output_filename,
        )

    plotter_util.output_subplot_array(
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )
    plt.close()


def potential_subplot(
    galaxy,
    grid,
    mask=None,
    positions=None,
    plot_in_kpc=False,
    figsize=None,
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    titlesize=10,
    xlabelsize=10,
    ylabelsize=10,
    xyticksize=10,
    mask_pointsize=10,
    position_pointsize=10,
    grid_pointsize=1,
    output_path=None,
    output_format="show",
    output_filename="galaxy_individual_potential",
):

    total_mass_profiles = len(galaxy.mass_profiles)
    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(
        number_subplots=total_mass_profiles
    )

    if figsize is None:
        figsize = figsize_tool

    plt.figure(figsize=figsize)

    unit_label, unit_conversion_factor = lens_plotter_util.get_unit_label_and_unit_conversion_factor(
        obj=galaxy, plot_in_kpc=plot_in_kpc
    )

    for i, mass_profile in enumerate(galaxy.mass_profiles):

        plt.subplot(rows, columns, i + 1)

        profile_plotters.potential(
            mass_profile=mass_profile,
            grid=grid,
            mask=mask,
            positions=positions,
            as_subplot=True,
            unit_label=unit_label,
            unit_conversion_factor=unit_conversion_factor,
            figsize=figsize,
            aspect=aspect,
            cmap=cmap,
            norm=norm,
            norm_min=norm_min,
            norm_max=norm_max,
            linthresh=linthresh,
            linscale=linscale,
            cb_ticksize=cb_ticksize,
            cb_fraction=cb_fraction,
            cb_pad=cb_pad,
            cb_tick_values=cb_tick_values,
            cb_tick_labels=cb_tick_labels,
            title="Galaxy Component",
            titlesize=titlesize,
            xlabelsize=xlabelsize,
            ylabelsize=ylabelsize,
            xyticksize=xyticksize,
            mask_pointsize=mask_pointsize,
            position_pointsize=position_pointsize,
            grid_pointsize=grid_pointsize,
            output_path=output_path,
            output_format=output_format,
            output_filename=output_filename,
        )

    plotter_util.output_subplot_array(
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )
    plt.close()


def deflections_y_subplot(
    galaxy,
    grid,
    mask=None,
    positions=None,
    plot_in_kpc=False,
    figsize=None,
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    titlesize=10,
    xlabelsize=10,
    ylabelsize=10,
    xyticksize=10,
    mask_pointsize=10,
    position_pointsize=10,
    grid_pointsize=1,
    output_path=None,
    output_format="show",
    output_filename="galaxy_individual_deflections_y",
):

    total_mass_profiles = len(galaxy.mass_profiles)
    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(
        number_subplots=total_mass_profiles
    )

    if figsize is None:
        figsize = figsize_tool

    plt.figure(figsize=figsize)

    unit_label, unit_conversion_factor = lens_plotter_util.get_unit_label_and_unit_conversion_factor(
        obj=galaxy, plot_in_kpc=plot_in_kpc
    )

    for i, mass_profile in enumerate(galaxy.mass_profiles):

        plt.subplot(rows, columns, i + 1)

        profile_plotters.deflections_y(
            mass_profile=mass_profile,
            grid=grid,
            mask=mask,
            positions=positions,
            as_subplot=True,
            unit_label=unit_label,
            unit_conversion_factor=unit_conversion_factor,
            figsize=figsize,
            aspect=aspect,
            cmap=cmap,
            norm=norm,
            norm_min=norm_min,
            norm_max=norm_max,
            linthresh=linthresh,
            linscale=linscale,
            cb_ticksize=cb_ticksize,
            cb_fraction=cb_fraction,
            cb_pad=cb_pad,
            cb_tick_values=cb_tick_values,
            cb_tick_labels=cb_tick_labels,
            title="Galaxy Component",
            titlesize=titlesize,
            xlabelsize=xlabelsize,
            ylabelsize=ylabelsize,
            xyticksize=xyticksize,
            mask_pointsize=mask_pointsize,
            position_pointsize=position_pointsize,
            grid_pointsize=grid_pointsize,
            output_path=output_path,
            output_format=output_format,
            output_filename=output_filename,
        )

    plotter_util.output_subplot_array(
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )
    plt.close()


def deflections_x_subplot(
    galaxy,
    grid,
    mask=None,
    positions=None,
    plot_in_kpc=False,
    figsize=None,
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    titlesize=10,
    xlabelsize=10,
    ylabelsize=10,
    xyticksize=10,
    mask_pointsize=10,
    position_pointsize=10,
    grid_pointsize=1,
    output_path=None,
    output_format="show",
    output_filename="galaxy_individual_deflections_x",
):

    total_mass_profiles = len(galaxy.mass_profiles)
    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(
        number_subplots=total_mass_profiles
    )

    if figsize is None:
        figsize = figsize_tool

    plt.figure(figsize=figsize)

    unit_label, unit_conversion_factor = lens_plotter_util.get_unit_label_and_unit_conversion_factor(
        obj=galaxy, plot_in_kpc=plot_in_kpc
    )

    for i, mass_profile in enumerate(galaxy.mass_profiles):

        plt.subplot(rows, columns, i + 1)

        profile_plotters.deflections_x(
            mass_profile=mass_profile,
            grid=grid,
            mask=mask,
            positions=positions,
            as_subplot=True,
            unit_label=unit_label,
            unit_conversion_factor=unit_conversion_factor,
            figsize=figsize,
            aspect=aspect,
            cmap=cmap,
            norm=norm,
            norm_min=norm_min,
            norm_max=norm_max,
            linthresh=linthresh,
            linscale=linscale,
            cb_ticksize=cb_ticksize,
            cb_fraction=cb_fraction,
            cb_pad=cb_pad,
            cb_tick_values=cb_tick_values,
            cb_tick_labels=cb_tick_labels,
            title="Galaxy Component",
            titlesize=titlesize,
            xlabelsize=xlabelsize,
            ylabelsize=ylabelsize,
            xyticksize=xyticksize,
            mask_pointsize=mask_pointsize,
            position_pointsize=position_pointsize,
            grid_pointsize=grid_pointsize,
            output_path=output_path,
            output_format=output_format,
            output_filename=output_filename,
        )

    plotter_util.output_subplot_array(
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )
    plt.close()
